<?php

/**
 * @file
 *
 * Admin functions to administer the predicate terms definitions.
 *
 * Split off from relationship.module for editing tidyness.
 *
 * Functions here relate only to the admin screens where
 * we can add and edit term definitions. None of the functions
 * in this file should be required to be loaded in a read-only
 * environment.
 *
 * The structure of this code was originally modelled on taxonomy.module,
 * and my RDF 'terms' are similar to Taxonomy 'terms',
 * I've discarded vocabulary support as being unneccessary, grouping is
 * handled by the ontologies.
 *
 * Terms can be defined and added to the term table,  preferably
 * via relationship_update_term();
 *
 * Pluggable Predicate Ontologies
 *
 * A module such as ontology_family may be created and enabled. It would
 * declare a bunch of predicates 
 * (IsFatherOf,  IsMotherOf,  IsParentOf, IsChildOf)
 * and their characteristics ( 
 *  IsMotherOf isSubPropertyOf IsParentOf,  
 *  IsChildOf isInverseOf IsChildOf
 *)
 * These predicates are added to the term list when the module is enabled,
 * and become available (on node edit) to anyone who wishes to use them. 
 * @see relationship_ontology_setup()
 *
 * To restrict or attach certain ontologies to only certain TYPES of node
 * (for example, you may create a custom node type for 'person', and only 
 * persons have family relations) the 'domain' of a term can be set.
 * This is analogous to associating node types with a vocabulary.
 *
 * Characteristics
 *
 * ... are axiomatic properties of predicates.
 *
 * Available characteristics should be used to relate one term with another.
 * Characteristics are modelled on the OWL definitions,  and include some
 * basic 'meanings' of term inter-relations. @see apply_characteristics.inc
 *
 * But sometimes that's not enough. 
 *
 * Predicate Ontology Modules can also define predicate logic lookups,
 * such as a routine that is called to calculate the results of a request to
 * 'IsCousinOf'.
 * This would return a meaningful result when the the system is asked for all
 * the cousins. This sort of lookup requres some coded,  repetative graph 
 * walking. If you wanted to encode the logic of a lookup HeirToTheThrone - 
 * this is where it would be done.
 *
 * These graph walks may be arbitrarily complex, and may need to call in 
 * many other inferred statements to get to their result. This on-demand
 * sort of query is different from the 'figure out what it means when we get it'
 * or on-save inferencing.
 *
 * The ontology libraries are also expected to provide appropriate ways of
 * displaying the relationships at node view time,  whether this be by publishing
 * a 'see also' sort of navigation block,  @see ontology_navigation.module 
 * or using hook_nodeapi() to modify the body output. 
 * @see ontology_revision_control.module
 * to see how isReplacedBy causes a warning to appear at the top of a page 
 * that is outdated.
 *
 * See the ontology modules for examples on how to make up
 * your own predicate programatically.
 *
 * 
 * Most of the time,  however the code treats 'terms' as objects,  
 * with a few properties and an array of characteristics.
 * Only when loading and serializing does this abstraction become important.
 *
 * Modules can also define their own characteristics and the
 * logical interpretation of what they imply,  however that 
 * is tricky and should be rare.
 *
 * @link predicates See Predicates for more about how I deal with predicate terms. @endlink
 * @see apply_characteristics.inc for how to code a characteristic.
 
 * @todo Use namespaces internally for indexing predicate terms? 
 * @todo be more selective
 *       how content types can choose various vocabularies to participate in.
 * ... partially done, using domain.
 *
 * @see ontologies
 * 
 * @ingroup docs
 * @package coders
 * @author Dan Morrison http://coders.co.nz/
 */
 
  
 
/**
 * @ingroup drupal_hooks
 * All these are pretty much boilerplate implimentations of stuff we've seen before.
 * @{
 */
 
/**
 * Implementation of hook_perm().
 */
function relationship_perm() {
  return array('administer relationships' , 'administer predicates', 'edit intrinsic properties');
}


/**
 * Implementation of hook_menu().
 *
 * As most of the admin screens are about administering predicates, 
 * the hook_menu is in this file.
 */
function relationship_menu($may_cache) {
  $items = array();

  if ($may_cache) {
    $items[] = array('path' => 'admin/relationship',  'title' => t('relationships'),
      'callback' => 'relationship_admin',
      'access' => user_access('administer relationships') );

    $items[] = array('path' => 'admin/relationship/list',  'title' => t('list predicates'),
      'access' => user_access('administer relationships'),
      'type' => MENU_DEFAULT_LOCAL_TASK,  'weight' => -10);

    $items[] = array('path' => 'admin/relationship/all',  'title' => t('list all'),
      'callback' => 'relationship_list_all',
      'access' => user_access('administer relationships'),
      'type' => MENU_LOCAL_TASK,  'weight' => -1);

    $items[] = array('path' => 'admin/relationship/add/term',  'title' => t('add term'),
      'callback' => 'relationship_admin',
      'access' => user_access('administer predicates'),
      'type' => MENU_LOCAL_TASK);

    $items[] = array('path' => 'admin/relationship/edit/term',  'title' => t('edit term'),
      'callback' => 'relationship_admin',
      'access' => user_access('administer predicates'),
      'type' => MENU_CALLBACK);

    $items[] = array('path' => 'relationship/edit',  'title' => t('relationship term'),
      'callback' => 'relationship_data_page',
      'access' => user_access('access content'),
      'type' => MENU_CALLBACK);

    $items[] = array('path' => 'admin/relationship/admin/tidy',  'title' => t('check database'),
      'callback' => 'relationship_check_database',
      'access' => user_access('administer predicates'),
      'type' => MENU_LOCAL_TASK);

  }

  return $items;
}


/**
 * Menu callback; dispatches to the proper relationship administration function.
 */
function relationship_admin() {
  $op = $_REQUEST['op'];
  $edit = $_POST['edit'];

  relationship_setup();

  if (empty($op) ) {
    $op = arg(2);
  }

  switch ($op) {
    case 'edit':
      if (arg(3) == 'term') {
        if(arg(4)){
          $output = _relationship_form_term( relationship_get_term( arg(4) , TRUE ) );
          break;
        }
        // else 'editing' nothing. DROP THROUGH to 'add'
      }
    case 'add':
      if (arg(3) == 'term') {
        $output = _relationship_form_term( array('type'=>CLASS_PROPERTY) );
      }
      break;
    case t('Delete'):
      if (!$edit['confirm']) {
        $output = _relationship_confirm_del_term(arg(4) );
        break;
      }
        // fall through:

    case t('Submit'):
      relationship_save_term($edit);
      relationship_flush_terms();
      drupal_goto('admin/relationship');
    default:
      $output = relationship_overview();
  }

  print theme('page',  $output);
}




/**
 * @}
 * End of normal hooks, continue with some module-specific functions.
 *
 * Routines for displaying the admin forms.
 */



/**
 * @ingroup relationship The Relationship Management Interface under Drupal
 * @{
 */

/**
 * List a summary of all terms and the actions we can do on them.
 * @returns html table
 */
function relationship_overview() {

  $op = $_REQUEST['op'];
  $edit = $_POST['edit'];

  if($op == t('Change Preferences')){
    foreach($edit as $param => $val){
      variable_set($param , $val);
    }
  }

  if($op == t('Hide Advanced Ontologies')){
    variable_set('relationship_show_advanced_terms',FALSE);
  }
  if($op == t('Show Advanced Ontologies')){
    variable_set('relationship_show_advanced_terms',TRUE);
  }

 
  $header = array(t('Name'),  t('Label'),  t('Characteristics'),   array('data' => t('Operations'),  'colspan' => '3') );

  // Start with all known predicates. Flush cache every visit to this screen, just for fun
  relationship_flush_terms();
  $terms = relationship_get_terms();
  
  // We'll list anything that's in an ontology
  $show_advanced = variable_get('relationship_show_advanced_terms',FALSE);
  
  // First enumerate the available, defined ontologies
  $ontologies = array();
  $ontology_statements = relationship_retrieve_statements_by_predicate(PREDICATE_TYPE, CLASS_ONTOLOGY);

  foreach($ontology_statements as $ontology_statement){
    $ontology_def = relationship_get_definition_by_name($ontology_statement['subject'] );
    $ontology_name = $ontology_def['uri'];
    
    $ontologies[$ontology_name] = array();

    // Skip the technical ones
    if(! $show_advanced && (($ontology_name == ONTOLOGY_RDF_AXIOMS)||($ontology_name == ONTOLOGY_OWL)) )
    { $skip = TRUE; } 
      else
    { $skip = FALSE;}
    
    if (! $skip){
      // prepare to display the block
      $data = $ontology_name;
      $rows[] = array( array('data' => '&nbsp;', 'colspan' => '5' ));
      $rows[] = array( array('data' => $data, 'colspan' => '5', 'style'=>'font-weight:bold'));
    }

    // Now retrieve all terms in this ontology, and format them
    debug("Listing terms defined in ontology '$ontology_name' ".( $skip?"(hidden)":"" ),2);
    $in_ontology_statements = relationship_retrieve_statements_by_predicate(PREDICATE_IS_PART_OF, $ontology_name );
    foreach($in_ontology_statements as $statement){
      $term = (object) relationship_get_definition_by_name( $statement['subject'], relationship_get_property_map('predicate') , TRUE );
      debug("Listing Term '$term->name'" ,3);
      if( $show_advanced || (! $skip && ( ((int)$term->weight) >= -5 ) ) ){
        $rows[] = _format_term_row($term);
      } # else {print('skipping <pre>'.print_r($term,1).'</pre>');}
      unset($terms[$term->name]); // Tick this off as done
    }
  }
  
  // Possible (but rare) that one of the original terms was not sorted into an ontology. 
  // List it anyway, so we can fix that. It'd be hard to find otherwise
  if($terms){
    $rows[] = array( array('data' => '&nbsp;', 'colspan' => '5' ));
    $rows[] = array( array('data' => 'Unsorted', 'colspan' => '5', 'style'=>'font-weight:bold'));
    foreach ($terms as $term) {
      if(!$term){continue;}
      $rows[] = _format_term_row($term);
    }
  }
  

  if (!$rows) {
    $rows[] = array(array('data' => t('No RDF terms available.'),  'colspan' => '5') );
  }


  $output = theme('table',  $header,  $rows);

  if(variable_get('relationship_show_advanced_terms',FALSE)){
    $prefs .= form_button(t('Hide Advanced Ontologies'));

    $prefs .= t("
      <p>
        The terms and classes shown here are all resources that are either of 
        <strong>type owl:Property</strong>  or anything that 
        <strong>isPartOf</strong> any groupings of <strong>type owl:Ontology</strong>.
      </p><p>
        The predicate terms and <em>characteristics</em> that show up here in 
        the summary page are those which are rated as <em>interesting</em>. 
        'Interesting' is based on weighting: <br/>
        Terms with negative weighting values do not show
        up in user select boxes. Terms with weight below -5 are so un-interesting
        they don't even show up on this page unless 'advanced' is turned on,
        <br/>
        But if you increase the weight of predicate characteristics 
        (properties of predicates) <em>over</em> 5 they are so interesting that
        they will show up summarized on this screen in the 'characteristics' column.
      </p>
    ");

  } else {
    $prefs .= t("<p>Terms that are in the built-in system ontologies (RDF Axioms and OWL) and terms that have weighting less than -5 (un-interesting) are currently <i>hidden</i> on this screen.<br/> %switch_button </p>", array('%switch_button' => form_button(t('Show Advanced Ontologies'))) );
  }
  $output .= form($prefs . l('More Settings','admin/settings/relationship') );

  $output .= t('
      The best way to define new terms is to install an ontology module which
      defines its own predicates, or install the <b>rdf_ontology.module</b> and
      import globally-defined terms from the web using that.
      <br/>Plugin Ontology modules should be available from the Drupal module repository.
  ');


  return $output;
}


/**
 * Format display of the defined characteristic properties.
 */
function _format_term_row($term) {
  $relationship = "";
  static $all_characteristics;
  if(! $all_characteristics){ $all_characteristics = relationship_get_all_characteristics() ;}
  
  foreach($all_characteristics as $name => $characteristic){
    // Only list INTERESTING characteristics here. 
    // Interesting means the weight of the predicate is > 4;
    if( $term->$name && (0 + $characteristic->weight) > 4 ){
      $relationship .= '<small class="term-characteristic"  >'.$name.($term->$name ? ":" . '</small><small class="term-characteristic-value">'.$term->$name : '').'</small><br/> ';
    }
  }
  

  $actions = l(t('edit term'),  "admin/relationship/edit/term/$term->name");
  $actions .= " " . l(t('delete'),  "admin/relationship/edit/term/$term->name",  NULL ,  "op=Delete");

  return array(
    ( $term->uri ? "<a href='{$term->uri}' title='Formal definition of this term'>{$term->name}</a>" : $term->name) ,
    '<acronym title="'.check_plain((string)$term->description).'" >'.check_plain((string)$term->label).'</acronym>',
    $relationship,
    $actions
  );
}

/**
 * Display an edit form for an RDF 'term'.
 * 
 * Set its name,  label,  characteristics and descriptions here.
 * Should also work for 'new' terms.
 * @param $edit assoc array of edit fields
 * @return HTML form
 */
function _relationship_form_term( $term_def = array() ) {
  
  $edit = object2array($term_def);
  
  $form = form_textfield(
    t('Predicate label'),  
    'label', 
    $edit['label'], 
    50, 
    64,  
    t('
        The human-readable predicate for this metadata predicate. 
        Use a descriptive phrase with leading caps. 
        Choose one from <a href="http://dublincore.org/documents/dcmi-terms/" 
        target="_new">the Dublin Core list</a> if possible. 
        Should be a descriptive sentence fragment that will make grammatical
        sense of the statement "Document A [....] Document B".  
        Example: "Is Part Of", "References", "Is Replaced By".
      '),  
    NULL,  
    TRUE
  );
      
  $form .= form_textfield(
    t('Predicate name'),  
    'name',
    $edit['name'],
    50,
    64,
    t('
        The internal label term. 
        May be the same as,  or a shorthand (CamelCase) form of the label. 
        Should match the <a href="http://dublincore.org/documents/dcmi-terms/" 
        target="_new">Dublin Core naming convention</a> if possible. 
        Example: "isPartOf", "References", "IsReplacedBy".
      '),  
     array('onclick' => "if(!this.value){this.value = this.form.elements['edit-label'].value.replace(' ','');}"),
     TRUE
   );
   
  $form .= form_textarea(
    t('Description'), 
    'description', 
    $edit['description'],
    60,
    5,  
    t('
        A description of what relationship the term is supposed to describe.
      '),
    array('onclick' => "if(!this.value){this.value = 'This resource '+this.form.elements['edit-label'].value+' the referenced resource.';}")

  ); 
      
  if( $term_def->derived ){  
    $form .= t('
        This property is <em>derived</em> which means it is never
        stated explicitly by a user, but is set only through deduction
        based on other relationships. Exactly what those relationships 
        <em>are</em> is defined in the characteristics below.
    ');
  }
      
  $form .= form_textfield(
    t('URI'),
    'uri',
    $edit['uri'],
    50, 
    64,  
    t('
        A reference to the definative resource (schema) that describes this term.
        A technical definition of what properties this metadata term implies. 
        Eg: 
          "http://purl.org/dc/terms/replaces", 
          "http://www.w3.org/TR/html4/types.html#type-links" . 
        Should probably be some implimentation of 
        <a href="http://www.w3.org/TR/owl-features/" title="Web Ontology Language">OWL</a>
        which can then define term properties (characteristics) like 
        "commutative" and "associative". 
        Or you can just make your own URI up or leave it blank.
        Even if its not a formal schema, it is still used as a namespace.
        Schemas can be found at 
        <a href="http://www.schemaweb.info" target="_new">Schemaweb</a>
        and other places.
      '),
      array('onclick' => "if(!this.value){this.value = this.form.elements['edit-name'].value;}"),
      FALSE
    );

  $type_urls = relationship_retrieve_subjects_by_predicate(PREDICATE_TYPE, CLASS_CLASS);
  foreach($type_urls as $url){
    $types[$url] = relationship_get_shortname($url);
  }


  // use longnames under the hood, display shortnames.
  $form .= form_select(
    t('Type'),
    'type',
    relationship_get_longname( $edit['type'] ) ,
    $types,  
    t('
        The type of resource this is. 
        Predicate Terms are of type <strong>Property</strong>
        but advanced options are possible if OWL classes are being defined
      ')
    );

  // Use the node relationship form here to apply statements to this form
  // Prepare the appropriate drop-downs 

  $characteristics = relationship_get_all_characteristics($edit['type']);
#  debug_pre(array_keys($characteristics));

  $predicates = array_merge( array(''=>''), relationship_get_terms() );

  // flatten this list - it's to be used as an option drop-down
  foreach($predicates as $i => $p){$predicate_list[$i] = $p->name;}
  
  $form .= relationship_form($term_def, $characteristics, $predicate_list);

  $form .= form_weight(
    t('Weight'),
    'weight',
    $edit['weight'],
    10,
    t('
        Importance for display and listings.<br/> 
        Terms with a negative weight are technical or implied, and will not show 
        up on user menus. For example, the term "rdf:isDefinedBy" is for internal
        use only, and is not a user-defined relationship. 
        Term "nav:last" would be an <em>implied</em> property only (derived from
        nav:next) and although it provides part of an ontology, it is not 
        something the user should override. 
        Term "creator" is liked to the Drupal internal "node->user" 
        property, and changing it could be dangerous.
        Giving all of these types of properties negative weighting will 
        hide them from users select boxes.<br/> 
        Characteristics can also be weighted - a value above zero makes it 
        "interesting" enough to appear on the predicate summary list.
    ')
  );


  $form .= form_button(t('Submit') );

  if ($edit['name']) {
    $form .= form_button(t('Delete') );
    $form .= form_hidden('name',  $edit['name']);
  }
  else {
    $form .= form_hidden('destination',  $_GET['q']);
  }

  //
  // Format the help about characteristics
  //
  $glossary="<p>
      There may be relationship <em>between</em> these relationship,  eg 
      '<strong>References</strong>' is the <em>inverse</em> of 
      '<strong>Is Referenced By</strong>',  and a '<strong>Is Part Of</strong>'
      property is <em>transitive</em> because it's implied that a part of a 
      part also a part of the whole.
  </p><p>
      For more detail,  refer to the 
      <a href='http://www.w3.org/TR/2004/REC-owl-guide-20040210/#PropertyCharacteristics'>
      Web Ontology Language</a>. 
      If used,  this extra information can be used as rules for 'graph walking'
      to deduce implicit metadata for items without stating them explicitly 
      each time,  eg a page can infer which pages reference it without 
      maintaining a list of referees itself.
  </p>\n";

  foreach($characteristics as $char_name => $char_def){
    $glossary.="\n<dt><a href='{$char_def->uri}' title='See the Formal definition of this characteristic'>{$char_def->name} : {$char_def->label}</a></dt>\n<dd>{$char_def->description}</dd>\n";
  }

  $glossary.='</dl>';
  
  $form .= form_group(t('About Property Characteristics'),  $form_group,  t($glossary) );


  /* Advanced diagnostic dump */  
  if($edit && variable_get('relationship_show_advanced_terms',FALSE) ){
    $dump = var_export($edit, 1);
    $form .= "<div><textarea rows='".(count(explode("\n",$dump))+2)."' cols='80'>".$dump."</textarea></div>";
  }

  if($edit['uri'] && function_exists('ontology_rdf_render_rdf_about') ){ // RDF interface is optional
    $form .= l(t('RDF'),  "relationship/rdf/".$edit['uri'],  
          array('title' => t('The pure RDF-metadata about this resource.') ) ); 
    $form .= ontology_rdf_render_rdf_about($edit['uri']);
  }
  
  return form($form);
}



 
/**
 * Save the input from an edit term form.
 *
 * Deal with a few special cases,  then serialize the data
 * @param $edit assoc array of edit fields
 * @returns assoc array,  possibly slightly modified for the serialization
 */
function relationship_save_term($edit) {

  if ($_REQUEST['op'] == t('Delete') ) {  return relationship_delete_term($edit['name']); }

  // need to read the flat form values back into statement structure
  $edit['statements'] = relationship_read_statements_from_form($edit);

  // Special-case the 'Inverse' property
  // tell the other relationship to point back to us
  if ($other_predicate_assertion = $edit['statements']['inverseOf'] ) {
    $other_predicate = $other_predicate_assertion['object'];
    drupal_set_message("As it's an inverse,  I'm linking target term $other_predicate back to this one as $characteristic also."); 
    $other_term = relationship_get_term($other_predicate);
    $other_term->characteristics[$characteristic] = $edit['name'];
    $other_term->inverseOf = $edit['name'];

    relationship_update_term($other_term);
  }

  if ($edit['name']) { 
    relationship_update_term(array2object($edit) );
    $message = t('The term %term has been updated.',  array('%term' => theme('placeholder',  $edit['name']) ) );
  }
  // remember to relationship_flush_terms(); if needed
  
  drupal_set_message($message);
  return $edit;
}

/**
 * Scan a form submission for rows of subject:predicate:object triples
 * and convert them into statement array format.
 *
 * Used as a utility for form_validate type actions.
 * @param $edit An array of field values, as submitted by an edit form.
 *              Expected to contain arrays subjects[] , predicates[] objects[]
 *              and optionally titles[]. 
 *              Horizontal rows in these vertical columns are our statements
 * @return array of statements.
 */

function relationship_read_statements_from_form(&$edit) {

 if (is_array($edit['predicates']) ) {
    $edit['statements'] = array();
    foreach ($edit['predicates'] as $count => $predicate) {
      if ($predicate) {
        $statement = array(
          'subject' => $edit['uri'],  
          'predicate' => $predicate,  
          'object' => trim($edit['objects'][$count]),
          'title'  => trim($edit['titles'][$count])
        );
        $edit['statements'][_relationship_hash_key($statement)] = $statement;
      }
    }
  }
  return $edit['statements'];
}

/**
 * @}
 */



/**
 * Show page warning about term deletion.
 * @param $name
 * @returns html
 */
function _relationship_confirm_del_term($name) {
  $term = relationship_get_term($name);

  $extra  = form_hidden('type',  'term');
  $extra .= form_hidden('name',  $name);

  $output = theme('confirm',
                  t('Are you sure you want to delete the term "%title" ?',  array('%title' => theme('placeholder',  $term->name) ) ),
                  'admin/relationship',
                  t('Deleting a term will delete all its defined relationship if there are any. This action cannot be undone.'),
                  t('Delete'),
                  t('Cancel'),
                  $extra);
  return $output;
}

/**
 * Hook into the core management of content types.
 *
 * Called from admin/node/configure/types
 * This displays the valid terms that can be used on the
 * given node type.
 *
 */
function relationship_content_type_settings($node) {

  $out .= t('Properties, annotations or relationships that can be added to nodes of this type:');

  $possible_ontologies = relationship_query($node->type,PREDICATE_CAN_USE_ONTOLOGY,NULL,'object');
  if( $possible_ontologies ){
    $out .= '<ul>';
    foreach($possible_ontologies as $ontology_name){
      $ontology_def = relationship_get_definition_by_name($ontology_name);
      $out .= '<li>Ontology: <b>'.$ontology_name.'</b>';
      $ontology_members = relationship_query(NULL, PREDICATE_IS_PART_OF, $ontology_name, 'subject');
      $out .= ' [' . join(', ',$ontology_members) . ']';
      $out .= '</li>';
    }
    $out .= '</ul>';
  }

  $possible_properties = relationship_query(NULL,PREDICATE_DOMAIN,$node->type,'subject');
  if( $possible_properties ){
    $out .= '<ul>';
    foreach($possible_properties as $property_name){
      $term_def = relationship_get_term($property_name);
      $out .= '<li>'.$term_def->label;
      $out .= ' [' . ( ($term_def->range)?$term_def->range:'no range defined' ) .']';
      $out .= '</li>';
    }
    $out .= '</ul>';
  }

  $out .= '<br/>'.l( 'Configure Relationships available to this node type', 'relationship/edit/'.$node->type);

  return form_group('Relationships',$out);
  
}

/**
 * Configuration form for associating valid properties with resources
 * (for administering classes - specifically node types)
 *
 */
function relationship_available_properties_form($class_name) {
  // Construct table exposing all predicates associated with this 
  // content type/Class
  
  require_once('coders_widgets/coders_widgets.inc');
  widgets_init('dynamic_table_rows');

  $predicates = relationship_get_terms();
  $predicate_list = relationship_terms_as_options($predicates);

  $new_button = '<input type="button" value="New" name="new_statement" id="new_button" '.
          'onClick="if (typeof(add_row_to_edit_table)==\'function\')'.
          '{return add_row_to_edit_table(this);}'.
          'else{alert(\'no dhtml function available\');return false;}" />';

  $fields = form_hidden('subjects][',$class_name) . 
      form_hidden('predicates][',PREDICATE_CAN_HAVE_PROPERTY).
      form_select('',  'objects][',  '',  $predicate_list  );

  $header = array('Available Relationships','');
  $row = array( $fields, $new_button );

  $out .= theme("table", $header, array($row), array('id'=>'metadata_entry_table') );

  return $out;  
}

/**
 * Performs a consistancy check on the database.
 *
 * Removes references to non-existant nodes
 * Allows a refresh of the predicate data.
 * Menu callback
 */
function relationship_check_database() {

  /* Clean up nids that don't exist */
  $out = "<p>
          <h2>Cleanup database</h2>
          Although it should not happen, it's possible that we could be left
          with references concerning nodes that no longer exist. This page can 
          check and clean these up. It's really just an admin function for
          development.
          </p>
  ";
  $query = "SELECT s.%s as nid FROM {relationship_statement} s LEFT JOIN {node} n ON ( s.%s = n.nid) WHERE s.%s > 0 AND n.nid IS NULL GROUP BY s.%s ";
  $misdirected_subjects = db_query($query,'subject','subject','subject','subject' );
  $out .= '<p>Found ' . db_num_rows($misdirected_subjects) . " statements in DB with an invalid subject<br/>";

  while($s = db_fetch_object($misdirected_subjects)){
    $out .= 'node:'.$s->nid.'<br/>'; 
    $out .= form_hidden('nid][',$s->subject);
  }
  $out .= '</p>';

  $misdirected_objects = db_query($query,'object','object','object' ,'object' );
  $out .= '<p>Found ' . db_num_rows($misdirected_objects) . " statements in DB with an invalid object<br/>";

  while($s = db_fetch_object($misdirected_objects)){
    $out .= 'node:'.$s->nid.'<br/>'; 
    $out .= form_hidden('nid][',$s->nid);
  }
  $out .= '</p>';
  $out .= form_button(t('Clean Up'));
  
  
  $op = $_REQUEST['op'];
  if($op == t('Clean Up')){
    $nids = $_REQUEST['edit']['nid'];
    drupal_set_message('Removing references to '.count($nids).' objects..');
    foreach ($nids as $nid ){
      relationship_resource_delete_statements($nid,TRUE);
    }
  }

  $out .= form_button(t('Drop'))." Totally wipe the relationship database";
  if($op == t('Drop')){
    drupal_set_message('Dropping ALL data from database');
    db_query('DELETE FROM {relationship_statement}');
  }


  $out .= "<p>
      <h2>Reset Axiomatic Predicates</h2>
      In the database, there are a set of underlying 'axioms' defining
      The very definition of the terms used as predicates. 
      These sometimes need tweaking. Make sure they are up-to-date.
  </p>";

  $out .= "<p>
      Things can (seem to) get very complicated very quickly.
      For this reason, many of the advanced features are not enabled out-of-the-box.
      Loading <strong>Advanced Axioms</strong> adds the concept of property inheiritance
      and domains and ranges - to restrict what type of property can apply
      to what type of subject, and what sort of object it will point to.
  </p>";

  $out .= form_button(t('Reload Axioms'));
  $out .= form_button(t('Load Advanced Axioms'));

  if($op == t('Reload Axioms')){
    relationship_import_definitions( relationship_get_axioms(), relationship_get_property_map('characteristic') );
  }
  if($op == t('Load Advanced Axioms')){
    relationship_import_definitions( relationship_get_more_axioms(), relationship_get_property_map('characteristic') );
  }
  
  /* Check remote URLs? */


  print( theme('page', form( $out ) ) );

}


/**
 * @ingroup relationship_api
 * Public Routines you might want to use from outside.
 * Mostly for adding, retrieving and updating term entries.
 * @{
 */

/**
 * Save changes made to a term object.
 *
 * The actual serialization. Change or create a term definition in the database.
 * relationship_save_term() is the UI form that usually calls this.
 * @param $term object definition
 * @return success
 * @ingroup relationship_api
 * @see relationship_delete_term()
 * @see predicates The predicate term object definition.
 */
function relationship_update_term(&$term) {
  // A term object is really an abstract collection of statements.
  // Saving the statements collectively means saving the term.

  if(! $term->name){ // This is an API, more paranoia
    drupal_set_message('Cannot save a term with no name identifier at all! Problem updating term.','error');
    return;
  }
  if(! $term->uri){ // This is an API, more paranoia
    $term->uri = $term->name;
    drupal_set_message(t('
        Setting term URI to just "%uri" - this means that the term is 
        basically only valid in the context of the current site. 
        This is OK for made-up relationships, but means this bit of data
        is harder to share with other servers, as it has no namespace.
    ', array('%uri' => $term->uri) ) );
  }


  $old_term = relationship_get_term($term->name);
  if($term->uri != $old_term->uri){
    // If they've changed the URI, it's annoying, as that's the UID for the term
    // Need to clear out everything that it used to know.
    // The update below will replace it with new values. Hopefully it won't break anything.
    relationship_delete_term($old_term->uri);
  }

  // Pack some of the top-level properties back into the term description as relationships.
  // - we don't save a term 'object', just a series of statements about a spurious term.
  // The actual name of the drupal object property may be remapped to a normative predicate label as we go
  // eg $thing->weight becomes a statement with predicate 'sortOrder'
  foreach(relationship_get_property_map('predicate') as $predicate => $property){
    // allow null case - which means don't serialize this property
    // Don't over-write it if it's already set
    if($predicate && (! $term->statements[$predicate] ) ){
      $term->statements[$predicate] = 
          relationship_statement($term->uri,  $predicate,  $term->$property); 
    }
  }
#debug_pre($term);

  // For convenience (but not efficiency) I'll try to use shortnames when saving statements
  // about my own ontologies. This will have side-effects later on.
  foreach($term->statements as $i => $s){
    relationship_validate_statement($term->statements[$i]);
  }
  
//  debug("<h2>Updating term</h2>");
//  debug_pre($term);
  
  return relationship_resource_save($term);
}

/**
 * Delete a term AND optionally all statements using that term.
 *
 * @param $name identifier of the term to remove, short or long version.
 * @param $extreme_prejudice annhilate all statements that USE this term as well. 
 * @return void
 * @see relationship_update_term()
 */
function relationship_delete_term($name, $extreme_prejudice=FALSE) {

  $uri = relationship_get_longname($name);
  relationship_resource_delete_statements( $uri );

  if ($extreme_prejudice) {
    db_query('DELETE FROM {relationship_statement} WHERE predicate = "%s" or predicate = "%s" ', $name, $uri  );
  }

  drupal_set_message(t('Deleted relationship term "%name" [%uri].',  array('%name' => $name, '%uri' => $uri) ) );

  relationship_flush_terms();
}





/**
 * Return a list describing all possible term characteristics.
 *
 * This list may be modified by plugins adding extra logic
 * and making appropriate term definitions claiming their 
 * new characteristic has 'domain' of 'property' or 'ObjectProperty'.
 *
 * As this list is pretty stable, it's cached. The cache should be flushed
 * whenever a term is modified, and it'll be rebuilt if empty.
 *
 * @param $flush if set, dump the cache and force a re-read, re-calculation of the data
 * @return assoc array of available characteristic definitions. 
 *         Indexed by the characteristic shortname.
 *         The definition should include the label and the URI of the characteristic
 * @see apply_characteristics.inc
 * @todo deprecate this, its a waste of time?
 */
function relationship_get_all_characteristics($type=CLASS_PROPERTY,$flush=FALSE) {
  $characteristic_cache_key = "relationship_characteristics";
  
# debug_pre($characteristics);

  if( ($cached = cache_get($characteristic_cache_key) ) && (!$flush)  ){
    return unserialize($cached->data); 
  }

  // Characteristics are those predicates whose 'domain' 
  // (things it applies to) is 'property'. 

  // $characteristic_list = relationship_query( NULL, PREDICATE_DOMAIN, CLASS_PROPERTY, 'subject' );
  // equivalent to:
  $characteristic_list = relationship_query($type, PREDICATE_CAN_HAVE_PROPERTY,NULL,'object');

  // thanks to inference, this list returns duplicates, due to the keying on full statement triples.
  // there may be more than one reason why the statement is returned, we've retrieved them ALL
  $characteristics = array();
  foreach($characteristic_list as $k=>$s){
    $s = relationship_get_longname($s);
    $characteristics[$s] = relationship_get_term($s);
  }

  cache_set($characteristic_cache_key,serialize($characteristics)) ;

  return $characteristics;
}


/**
 * @}
 */

/**
 * Implimentation of (proposed) hook_setup() .
 *
 * Insert (or remove) a set of term definitions into the predicate table.
 * Called once on first-run,  or optionally with a 'refresh' flag to force a reversion.
 * 
 * This is a helper utility function for use by pluggable ontologies. If an
 * ontology declares an ontology_defined_terms() sort of function,  which returns 
 * an array of term defs,  AND this pluggable ontology has just been enabled,
 * insert those term definitions into the datastore for use.
 * Or,  if the ontology has just been disabled (through the use of modules admin)
 * remove the terms. This is non-destructive,  and leaves any defined relationships
 * behind,  although their logic will probably not work any more.
 *
 * Internally,  we set a variable flag {modulename}_db_initialized to monitor
 * the installation status of a module.
 *
 * @param $module_name
 * @param $refresh Force the installation, even if it's apparently already there.
 * @ingroup ontologies
 */
function relationship_ontology_setup($module_name,  $refresh=0) {
  // I'm passing the module name rather than just the terms array itself
  // in case I want to do anything more advanced with this function some time.

  $get_terms = $module_name.'_defined_terms';
  // if the module is turned on,  but not ready for action, 
  // make sure the db lists its published predicates.
  if (function_exists($get_terms) ) {
    if (module_exist($module_name) && ((! variable_get($module_name.'_db_initialized', FALSE) ) || $refresh) ) {
      
      drupal_set_message( t('Setting up new module %module_name', array('%module_name' => $module_name)) );

      $new_terms = $get_terms();
      
      $import_template = relationship_get_property_map('predicate');
      relationship_import_definitions( $new_terms , $import_template);

      drupal_set_message("Initialized node relationship (predicate) terms used by $module_name. The predicate_list should now include ".implode(",  ", array_keys($get_terms() ) ) );
      variable_set($module_name.'_db_initialized', TRUE);
    } 
    else 
    {
      if (variable_get($module_name.'_db_initialized', FALSE) && (! module_exist($module_name) ) ) {
        // We've just been turned off. Clean up a little. Remove terms from database. 
        // Leave relationships using that term however. 
        // This may orphan them but it's better than data loss.
        foreach ($get_terms() as $term_name => $term_def)
        {
          relationship_delete_term( $term_name  );
        }
        relationship_get_terms('', TRUE); // this will flush the terms cache for us

        variable_del($module_name.'_db_initialized');
        drupal_set_message("
            Disabled the node relationship (predicate) terms used by 
            $module_name. The terms ".
            implode(",  ", array_keys($get_terms() ) )." 
            Have been disabled,  however any statements still using those terms
            were left behind on the nodes. 
            To vaporize them totally,  you would have been best to DELETE the 
            term using the term administration interface - if that is your intent.
        ");
      }
    }
  } // finished updating terms definitions.


}


