<?php

/**
 * @file
 *  
 * Supports the node-specific APIs and hooks for relationship
 * Renders the relationship table,  validates and saves links against nodes.
 *
 * A node can now have any number of 'statements' attached to it.
 * A 'statement' is an RDF 'triple' relating one resource with another.
 * EG,  the "next" resource from page one is page two.
 *
 * Simple and complex stuctures can be built from a collection of such
 * relationships.
 * 
 * Primary statements behave as if they are attached directly to the
 * subject - the node. They are exposed for editing in the node edit form
 * and should be saved alongside it. Secondary statements are inferred
 * from logical patterns,  and although they may get stored in the node
 * as $node->relationships during processing and rendering,
 * they should not be saved unless you want every logical relationship 
 * to become explicit.
 *
 * This file provides the actual relationship editing UI form, and not much
 * else. To this end it uses the range_selectors.inc and combobox.js
 * to provide the dynamic UI.
 *
 * @package coders
 * @author Dan Morrison http://coders.co.nz/
 */


// we need to filter out some of the predicates, both for security
// and to keep the list managable.
// These constants describe what we show

/**
 * Intrinsic values are those predicates that have a direct mapping with
 * Drupal internals. Need to control access to this group.
 * An intrinsic property is one that has a drupalEquiv mapping value set.
 */
define('RELATIONSHIP_FILTER_INTRINSIC_VALUES',1);
/**
 * Trivial values are predicates that have an importance/weight less than zero.
 * This includes predicates that are only ever implied. They should be not
 * show up on user edit forms
 */
define('RELATIONSHIP_FILTER_TRIVIAL_VALUES',2);



/** @defgroup admin 
 *  
 * Administrator Interface
 * Hooks and functions for administering node relationships.
 * @{
 */


/**
 * Implementation of hook_nodeapi().
 *
 * based on taxonomy_nodeapi()
 */
function relationship_nodeapi(&$node,  $op,  $teaser = NULL,  $page=NULL) {
  //drupal_set_message("Calling relationship_nodeapi('$op' ".$node->nid);
  
  //if (!$node->nid) {return;} // can't do nothing with nothing
  $active_node_types = variable_get('relationship_nodes',array());
  //drupal_set_message("relationships apply to ".print_r($active_node_types,1));
  if (! in_array($node->type, $active_node_types)) {return;} // Not a node type we deal with
  
  switch ($op) {
    case 'form post':
      return relationship_form($node);
      break;
    case 'validate':
      return relationship_resource_validate($node);
      break;
    case 'insert':
      relationship_resource_save($node);
      break;
    case 'update':
      relationship_resource_save($node);
      break;
    case 'delete':
      
      // See if any relationships imply that this deletion should have a 
      // further-reaching effect. If this item isVampyreKingOf something,
      // then destroying this means that should be destroyed too...
      // Generally speaking, hasPart would exhibit this behaviour, but take care...
      if ($statements = relationship_query($node->nid, NULL, 1) )
      {
        foreach ($statements as $statement)
        {
          $term_def = relationship_get_term($statement['predicate']);
          // TODO
          // Find a way to abstract this operation 
          // - aren't there other times we'd want to use predicate hooks?
          if ($term_def->characteristics && isset($term_def->CascadeDeletion ) )
          {
            drupal_set_message("Cascading this action '$op' to other node ".$statement['object']);
            node_delete(array('nid' => nid($statement['object']) ,  'confirm' => 1) );
          }
        }
      }
    
        // It seems that the aliases have already been deleted before this runs.
        // How am I supposed to clean this up?
        // statements made using the alias as the subject will be left behind.
      //  foreach( relationship_get_aliases($node->nid) as $alias){
      //  relationship_resource_delete_statements($alias);
      //  }
      relationship_resource_delete_statements($node->nid);
      
      break;

    case 'load':
      $subject = ($node->nid) ? $node->nid : $_REQUEST['q'];

      // Also extrapolate any useful built-in node data into statement form
      // This may seem a lot of work, but is used for meta tags and RDF
      $intrinsic_node_statements = relationship_array2statements( (array)$node, $subject, relationship_get_property_map('node') , TRUE );
      
      // flag each of these statements as derived, so they don't get changed inappropriately
      foreach($intrinsic_node_statements as $i => $s){
        $node_statements[$i]['depth'] =RELATIONSHIP_IS_CALCULATED;
        $node_statements[$i]['why'] = 'intrinsic property from node_load';
      }

      // Careful not to add anything (even an empty array) if blank,
      // It messes up other folks logic when testing if the node exists.
      if ( ($subject && ($statements = relationship_query($subject) )) || $intrinsic_node_statements ) {
        $node->statements = $statements;
        $node->statements = array_merge($node->statements, $intrinsic_node_statements); // drupal data takes precedence if there's any conflict
      }

      break;

    case 'view':

      if ($page) {
        // Insert all our relationship links into the header,  to make the semantic web happen!
        if (! isset($node->statements)){
           relationship_nodeapi($node, 'load');
           debug("This shouldn't get here - legacy leftover. The node has no statements. Since when do we view before loading?");
        }
        relationship_add_statements_to_head($node->statements);
      } // page?

    break; // case view

    case 'settings': 

      // Integration with Drupal core - extend node types with extra predicates.
      // Add a list of the available predicates
      // (properties) to the configure content type page
      return relationship_content_type_settings($node);
    
  } // switch
}

/**
 * Generate a form for displaying and selecting relationship applying to a node.
 *
 * The form is supported by some DHTML on the client. @see coders_widgets.inc
 *
 * @param $node object
 * @param $predicates optional list of predicates to be used in the relationship drop-down
 * @param $objects optional list of urls to be used in the relationship object drop-down
 *
 * @returns HTML table form for adding metadata statements
 *
 * @todo significant rewrite.
 */
function relationship_form(&$node, $predicates=NULL, $objects=NULL) {

  // Enable combo box functionality
  require_once('coders_widgets/coders_widgets.inc');
  widgets_init('combobox');
  widgets_init('paired_selects');
  widgets_init('editable_hidden_fields');
  widgets_init('dynamic_table_rows');
  

  if (!$node->statements) { 
    // Should have already been loaded, but..
    if ($node->nid) {
      // fetch existing statements
      $node->statements = relationship_query($node->nid);
    }
    else {
      // new, empty resource
      $node->statements = array();
    }
  }

  // May need to unpack existing statements into the flat array,  so as
  // to expose all the fields to a form... the inverse of what validate() does
  if (is_array($node->statements) ) {
    $node->subjects   = array(); // due to alias support, we may deal with subtly different current subjects
    $node->predicates = array();
    $node->objects    = array();
    $node->titles     = array();

    foreach ($node->statements as $count => $statement) {
      $node->subjects[]   = $statement['subject'];
      $node->predicates[] = $statement['predicate'];
      $node->objects[]    = $statement['object'];
      $node->titles[]     = $statement['title'];
    }
  }


  # drupal_set_message("Displaying relationship statements <pre>".print_r($node, 1) ."</pre>");

  $header = array('this', 'relates to', '', array('data'=>'that','width'=>'100%'), '');

  // List all existing data in a table here
  // There are a few extra ids and things in here for javascript widgets to do their job
  $relationship_terms = relationship_get_terms();
  
  $show_trivial = variable_get('relationship_show_advanced_terms',FALSE);
  $show_intrinsic = variable_get('relationship_show_intrinsic_terms',FALSE);
  $can_edit_intrinsic = user_access('edit intrinsic properties');

  if(is_array($node->statements)){
    foreach ($node->statements as $i => $statement) {

      // TODO this routine needs to be refactored - it was constructed sorta ad-hoc
      // logic is currently broken down into the tiniest pieces, to relly lay out whats happening
      
      $s_cell = $o_cell = $p_cell = $c_cell = $title = $title_atts = $object_atts = '';

      $is_trivial = ($statement['depth'] < 0);
      $is_explicit = ($statement['depth'] == 0);
      $is_deduced = ($statement['depth'] > 0);

      if( $is_trivial && ! $show_trivial){continue;}

      // One day I should put fine-tuned authentication here.
      // It's reasonable to set permissions on a predicate-by-predicate basis.
      
      $term_def = $relationship_terms[$statement['predicate']];
      $is_intrinsic = $term_def->drupalEquivalent;

      if( $is_intrinsic && ! $show_intrinsic){continue;}

      $can_edit = TRUE;
      if( $is_intrinsic && ! $can_edit_intrinsic ){ $can_edit = FALSE;}
      if( $is_deduced ){ $can_edit = FALSE; }

      // if a predicate has no defined range (eg 'Node' or 'Resource')
      // treat it as a literal ... UNLESS it already looks like an URL
      // these are just guesses to make up for incomplete ontologies.
      // Having no defined range is bad.
      $is_literal = (($term_def->range == CLASS_LITERAL) || (!$term_def->range && !is_uri($statement['object']) && !is_numeric($statement['object'])) );


      // Display subject only if it's different than expected (an alias)
      $s_cell = ($node->nid && ($statement['subject'] != $node->nid))
                ?  "<span class='subject'>{$statement['subject']}</span>" 
                : '';

      $p_cell = "<em class='predicate' title='{$term_def->description}'>". 
        (($term_def)? $term_def->label : $statement['predicate']) ."</em>";

      $c_cell = '';

      if($can_edit){
        // Statement is partially or fully editable
        
        // need hidden fields
        $p_cell .= form_hidden('predicates][',  $statement['predicate']);
        $s_cell .= form_hidden('subjects][',  $statement['subject']);
        $o_cell .= form_hidden('depths][',$statement['depth'] );
      

        $object_atts = array('id' => 'hidden_object_'.$i );
        $link_atts = array('id' => 'exposed_object_'.$i);
        
        // can rename the title
        if ($is_explicit ){
          if(!$is_literal){
            // Only explicit links can be titled
            // Only save titles that are different from the default. 
            // No need to fill up the info with redundancies, and they have a disturbing habit
            // of getting out of date.
            $title = ($statement['title'] != $statement['object'])?$statement['title']:'';
            $title_atts = array('id' => 'hidden_'.$i, 'class' => 'changeable' , title => 'Label:');
            $link_atts = array('id' => 'exposed_'.$i);
          } else {
            $object_atts = array('id' => 'hidden_object_'.$i, 'class' => 'changeable' , title => 'Value:');
          }
        }

        $o_cell .= form_hidden_atts('objects][',$statement['object'] , $object_atts );
        $o_cell .= relationship_make_link( $statement, $link_atts );
        $o_cell .= form_hidden_atts('titles][', $title  , $title_atts ); // 

        if (! $is_literal ){
          // Trigger possible
          $c_cell = '<input type="button" value=">" '.
            ' onClick="change_select_range_to(\'hidden_object_'.$i.'\',\''.$statement['predicate'].'\'); init_combo_box_named(\'hidden_object_'.$i.'\') ; remove_element_named(\'exposed_'.$i.'\') " title="Choose from other valid values for this property" />';
        }


        if ($is_explicit ){
          // Only Explicit statements can be deleted
          // Writing button by hand as the drupal render is too bulky to fit in a table
          // This primarily works as javascript, but falls back as a submit action on noscript
          $action_cell = '<input type="submit" value="Del" name="edit[delete_statement]['.$i.']" '.
            ' onClick="if (typeof(delete_row_from_edit_table)==\'function\') {delete_row_from_edit_table(this); void(1);}" />';
        }

        if ($is_trivial ) {
            $action_cell = "<small title='This actual statement is derived from data stored internally or elsewhere '>implicit</small>";
        }

      } else {
        // can't edit this
        $o_cell = relationship_make_link( $statement );

        if( $is_deduced ){
          $action_cell = "<small title='This actual statement is in another page,  or due to the logical structure of the documents. It cannot be directly edited from this page '>... by implication</small>";

        }
        if( $is_implicit ){
          $action_cell = "<small title='This value is handled by the Drupal internals, or another module.'>built-in</small>";
        }
          // And can't do much more than that with this statement
      }

      $rows[] =array(
        'subject'   => $s_cell ,  
        'predicate' => $p_cell ,
        'change'    => $c_cell ,
        'object'    => $o_cell,
        'action'    => $action_cell
     );
     
    } // each statement
  } // has existing statements
  

  // Append a new editable row for a new statement 
  // Initialize javascript selectbox updater
  include_once("range_selectors.inc");

  // Figure what predicates are appropriate to put into the select,
  $options_filter = 0;
  if(! ( $show_intrinsic && $can_edit_intrinsic )){
    $options_filter = RELATIONSHIP_FILTER_INTRINSIC_VALUES;
  }
  
  #debug_pre($predicates);
  
  if(! $predicates){
    // This lookups should fetch not just explicitly defined
    // properties, but also others according to the lookup rules in 
    // relationship_lookup_CanHaveProperty()
    // todo - relationship_lookup_CanHaveProperty() needs work
    $predicate_list = relationship_query($node->type, PREDICATE_CAN_HAVE_PROPERTY,NULL,'object');
    $predicates = array_map(relationship_get_term, $predicate_list);

    #$predicate_list = relationship_query($node->type, PREDICATE_CAN_HAVE_PROPERTY,NULL);
    #debug("Statements about which predicates are appropriate to display on this ".$node->type." form are:<pre>".print_r($predicate_list,1)."</pre>" );

    # debug_pre($predicates);
  
    $options_filter = ($options_filter | RELATIONSHIP_FILTER_TRIVIAL_VALUES);
  }


  // flatten the definitions into a (restricted) option list
  $predicate_list = relationship_terms_as_options($predicates,$options_filter);

  # debug("Predicates appropriate to display on this ".$node->type." form are:<pre>".print_r($predicates,1)."</pre>" );

  $predicate_list = array_merge(array(""=>""), $predicate_list ); //insert blank first field 
  
  $output .= range_selectors($predicates, 'predicates][', 'objects][' );

  // Prime the object selectbox
  if(! $object_list ){
    $object_list = relationship_all_nodes_as_options();
  }
  
  $rows[] = _relationship_statement_fields( $predicate_list, $object_list );
  
  $output .= theme("table", $header, $rows, array('width'=>'90%', 'id'=>'metadata_entry_table') );

  if( variable_get('relationship_show_intrinsic_terms',FALSE) ){
    $output .= t('<p><small>You are viewing all known intrinsic properties of the node, as well as the explicitly stated relationships. Making changes directly to some of these <i>may</i> interact strangely with other modules.</small></p>');
  }

  $output =  form_group(t('Relationships and Metadata'),  $output)."\n<br/>";


  return $output;
}



/**
 * Convert a list of term def objects into an array
 * suitable for using in a select box.
 *
 * They will be indexed either of shortname (name) or longname (uri) depending on global settings
 *
 * Checks that only resources of type 'property' show up,
 * as relationship_lookup_CanHaveProperty() is currently a bit inaccurate
 * 
 * @param The current list or subset of predicate terms definitions.
 * @param $filter A boolean combination of options indicating which of the
 *        available terms NOT to show. 
 *        By default both filters are on - you must choose to turn them off.
 * @see RELATIONSHIP_FILTER_INTRINSIC_VALUES
 * @see RELATIONSHIP_FILTER_TRIVIAL_VALUES
 * @see drupal_get_terms()
 */
function relationship_terms_as_options($terms , $filter = 3 ) {
  if(! is_array($terms)){
    trigger_error("Invalid arg '$terms' to relationship_terms_as_options() ",E_USER_WARNING);
    return array();
  }
  $options = array();

  // index this list on long or short names
  $relationship_long_names = variable_get('relationship_long_names', FALSE);
  
  foreach($terms as $i => $term){

    // Skip the low-level terms - they are too trivial to list
    if(( $filter & RELATIONSHIP_FILTER_TRIVIAL_VALUES ) && ($term->weight < 0)){ continue; }

    // Do not list built-in properties unless given special permission
    if(( $filter & RELATIONSHIP_FILTER_INTRINSIC_VALUES ) && ($term->drupalEquivalent)){ continue; }

    if($term->type != CLASS_PROPERTY ){ continue; }

    // @TODO greater granularity of access can be applied here
    // else
    $indexid = ($relationship_long_names)?$term->uri:$term->name;
    $options[$indexid] = $term->label;
  }
  return $options;
}


/**
 * Read an edit form into a set of statements.
 *
 * Validating a nodes statements means reading in the 'predicate' and 'object'
 * arrays from the 'edit' form and storing them as 'statements' in the node.
 * Nodeapi callback
 * @param $node object
 * @return node object
 */
function relationship_resource_validate(&$node) {
  debug("Validating input ".print_r($node, 1),3 );

  if (is_array($node->predicates) ) {
    $node->statements = array();
    foreach ($node->predicates as $count => $predicate) {
      if ($node->delete_statement[$count]) {
        // this is for noscript support
        drupal_set_message("Discarding statement $count");
      } else {
        if ($predicate) {
          $subject = ($node->subjects[$count])? $node->subjects[$count] : $node->uri; 
          $node->statements[] = array(
            'subject' => trim($subject),  
            'predicate' => $predicate,  
            'object' => trim($node->objects[$count]),
            'title'  => trim($node->titles[$count]),
            'depth'  => $node->depths[$count],
            'why'    => ($node->whys[$count])?$node->whys[$count]:'user/'.$GLOBALS['user']->uid
         );
        }
      }
    }
    // clean up. probably unneeded, but less confusing in data dumps
    unset($node->subjects);
    unset($node->predicates);
    unset($node->objects);
    unset($node->titles);
    unset($node->depths);
  }

  // If given, also allow relationship statements to be absorbed over top of
  // intrinsic node properties. This gives low-level access to the guts of a
  // node (probably a bad thing)
  $data_cloud = relationship_statements2array($node->statements, relationship_get_property_map('node') );
  foreach($data_cloud as $predicate => $object){
    if( isset($node->$predicate) &&  ( $node->$predicate != $object) ){

      // Filter for user-access restrictions. Don't want to expose Drupal guts randomly
      if( user_access('edit intrinsic properties') ){
        debug("Over-writing node $predicate = {$node->$predicate} with $object ",2);
        $node->$predicate = $object;
      } else {
        drupal_set_message(t('You don\'t have permissions to modify the %predicate directly on this resource, but thanks for trying!',array('%predicate'=>$predicate) ));
      }
    }
  }
//debug('In');
//debug_pre($node);

  // init title if needed
  relationship_get_titles($node->statements, TRUE);
  return $node;
}




/**
 * Save the relationship associations (statements) of a given resource.
 *
 * AKA relationship_node_save() if following drupal module conventions.
 *
 * The resource object handed in (usually a node) is expected to have an array
 * called 'statements' attached to it. 
 * Each of these statements are serialized into the database.
 * @see statements for the structure of a 'statement'.
 *
 * If it wasn't explicitly set in the statement, 
 * the 'subject' is assumed to be this node.
 * Statements can have an optional title.
 *
 * $statements = array(
 *   array('predicate'=>'IndexedBy',  'object'=>'sitemap'),
 *   array('predicate'=>'LinksTo',  'object'=>'sitemap'),
 *   array('predicate'=>'LinksTo',  'object'=>'homepage', 'title' = 'Home'),
 *);
 *
 * As the method is currently 'wipe the old and replace the new'
 * any statement IDs are lost on each new save.
 * This is not a problem yet... still not actually using rsids.
 *
 * @param $resource An object containing an array of statements
 * @return void
 */
function relationship_resource_save(&$resource) {
  if ( (! $uri = $resource->nid ) && (! $uri = $resource->uri ) ) {
    drupal_set_message("Can't save metadata with no node ID or URI"); return;
  }
  
  if (is_array($resource->statements) ) {
    // careful not to wipe statements if the nodes data is simply undefined, 
    // maybe we didn't load the full set in the first place.
    relationship_resource_delete_statements($uri);
    foreach ($resource->statements as $statement) {
      if (is_array($statement) ) {
        if (! $statement['subject']) {
          $statement['subject'] = $uri; 
          // brand new nodes may not have had the nid available at statement time.
        }
        relationship_save_statement($statement,"Saving Resource");
      }
    }
  }
}





/**
 * Generate form elements for defining relationships.
 *
 * subject (current node, implied from context)
 * predicate (relationship term)
 * object (freeform,  probably a node id or URL)
 *
 * Return as an array so the respective parts can be
 * dealt with separately (put into a table row)
 *
 * Helper for node_form
 *
 * @param $predicates optional list of predicates to be used in the relationship drop-down
 *                    if not defined, ALL things of type 'Property' will be listed
 * @param $objects optional list of titles=>uris to be used in the relationship object drop-down
 * @returns an array to be put into a table row.
 * @todo test noscript
 */
function _relationship_statement_fields($predicate_list=NULL, $object_list=NULL) {

  // the library will supply the needed javascript
  if(function_exists('range_selectors_trigger_update')){
    $predicate_select_extra = range_selectors_trigger_update('predicates][','objects][');
  }
  
  return array(
    'subject' => 'this' ,  
    'predicate' => array('data' => form_select('',  'predicates][',  '',  $predicate_list, NULL ,$predicate_select_extra ), 'colspan'=>'2' ),
    'object' => form_select('',  'objects][',  '',  $object_list  ,  NULL,  ' class="combo" width="70%" '),
    'action' => '<input type="button" value="New" name="new_statement" id="new_button" '.
      'onClick="if (typeof(add_row_to_edit_table)==\'function\')'.
      '{return add_row_to_edit_table(this);}'.
      'else{alert(\'no dhtml function available\');return false;}" />' 
 );
}

/**
 * Return a list of ALL node items as an option list.
 *
 * Could get really long.
 * Uses aliases wherever possible,  otherwise returns links
 * just as integer nids
 *
 * @todo improve this part of the interface
 */
function relationship_all_nodes_as_options() {
  $aliases = array();
  $alias_lookup =  db_query("SELECT n.nid AS link ,  a.dst FROM {node} n INNER JOIN {url_alias} a ON CONCAT('node/', n.nid) = a.src;");
  while ($row = db_fetch_array($alias_lookup) ) {$aliases[$row['link']] = $row['dst']; }

  $options = array();
  $options[] = ""; // start with a blank
  $node_lookup = db_query("SELECT n.nid AS link,  n.title FROM {node} n ");
  while ($row = db_fetch_array($node_lookup) ) {
    if ($aliases[$row['link']]) {$row['link'] = $aliases[$row['link']];}
    $options[$row['link']] = $row['title'];
  }

  return $options;
}


/**
 * A query lookup callback, returns all predicates that may be available to a type.
 *
 * All the properties a subject can have!
 * This includes a logical lookup not implied by the ontology.
 * These callback lookups are where we CAN put business logic that we can't
 * yet express in statements.
 *
 * In addition to those properties added explicitly in the datastore, we should return 
 * - all properties from any associated ontologies, 
 * - all properties that may apply to the superclass (node)
 * - all wildcard properties that have NO domain defined.
 *
 * @TODO check that members of ontologies really are 'properties'
 *       I'm getting a few false properties due to classes, non-node properties
 *       and ontologies being inside ontologies too.
 *       I don't really want EVERYTHING in an ontology.
 * @ingroup ontologies
 */
function relationship_lookup_CanHaveProperty($subject){

  $statements = array();
  $possible_ontologies = relationship_query($subject,PREDICATE_CAN_USE_ONTOLOGY,NULL,'object');
  foreach($possible_ontologies as $ontology_name){
    // for all ontologies this class participates in...
    $ontology_statements = relationship_query(NULL, PREDICATE_IS_PART_OF, $ontology_name);
    // each member of that ontology
    foreach($ontology_statements as $statement){
      // Note - that was a reverse lookup, so we use the 'subject' as our result
      // adjust all the derived statements to refer to the current subject
      $s = relationship_statement( $subject, PREDICATE_CAN_HAVE_PROPERTY , relationship_get_shortname( $statement['subject'] ) );
      $s['depth'] = (int)$statement['depth'] + 1;
      $s['why'] = $statement['why']."; in ontology $ontology_name, which is associated with ".$subject;
      $statements[_relationship_hash_key($s)] = $s;
    }
  }
 
  // all properties that may apply to the superclass (node)
  // todo - lookup the actual superclass, don't assume node
  if($subject != CLASS_NODE){ // avoid loop
    $superclass_statements = relationship_query(CLASS_NODE, PREDICATE_CAN_HAVE_PROPERTY, NULL);
    // adjust all the derived statements to refer to the current subject
    foreach($superclass_statements as $statement){
      $s = relationship_statement( $subject, PREDICATE_CAN_HAVE_PROPERTY , relationship_get_shortname($statement['object']) );
      $s['depth'] = (int)$statement['depth'] + 1;
      $s['why'] = $statement['why']."; in ontology $ontology_name, which is associated with class: ".CLASS_NODE;
      $statements[_relationship_hash_key($s)] = $s;
    }
  }


  #debug_pre(array_keys($statements));

  // all wildcard properties that have no domain defined.
  foreach(relationship_get_terms() as $term){
    if(!$term->domain){
      $s = relationship_statement( $subject, PREDICATE_CAN_HAVE_PROPERTY , $term->name );
      $s['depth'] =  RELATIONSHIP_IS_CALCULATED;
      $s['why'] = "Its a predicate term with no domain limit";
      $statements[_relationship_hash_key($s)] = $s;
    }
  }

  #debug_pre(array_keys($statements));

  return $statements;  
}


/**
 * List everything about everything. Page.
 *
 * Menu callback.
 * Admin function / diagnostic / scaffolding
 * Should really use pager
 */
function relationship_list_all() {

  drupal_set_title('Dumping all relationships in the database table');
  $output = 'Debug info only - this will get too big very quickly';
  
  $all_relationship_lookup = db_query("SELECT * FROM {relationship_statement} order by subject;");
  $rows = array();
  while ($statement = db_fetch_array($all_relationship_lookup) ) {
    $row = $statement;
    $row['subject'] = l( scrunch_url( $statement['subject'] ), 'relationship/edit', NULL, 'edit[uri]='.urlencode($statement['subject']) );
    unset($row['title']);
    $rows[] = $row;
    $headers = array_keys($statement);
  }
  unset($headers['title']);
  $output .= theme('table', $headers  , $rows);

  print theme('page',$output);  
}



/**
 * Expose all statements about any subject.
 *
 * Menu callback
 *
 * The target subject can either be handed in as the remainder or the URL
 * or explicitly as a 'uri' form parameter.
 * This is a generic edit form for a resource. A resource is represented as
 * a collection of statements about the same thing.
 *
 * @see relationship_list_all()
 * @todo make this a saving edit form
 */
function relationship_data_page() {
  $op = $_REQUEST['op'] ? $_REQUEST['op'] : arg(1);
  $subject = arg(2) ? arg(2) : $_REQUEST['edit']['uri'] ;
  $edit = $_REQUEST['edit'];
  
  drupal_set_title("All info about '$subject'");

  if(!$subject){
    drupal_set_message(t('No subject defined'),'error');    
  }

  if($op == t('edit')){
    // Entering from somewhere else
    $resource = relationship_get_definition_by_name($subject, NULL, TRUE);
    if(! $resource['uri'] ){
      // brand new
      $resource['uri'] = $subject;
    }
    $resource = (object)$resource ;
  } else {
    // load submission data (if any)
    $resource = (object)$edit;
  }

  relationship_resource_validate($resource);

  if($op == t('Submit')){
    relationship_resource_save($resource);
  }

  $out .= form_textfield( t('URI'), 'uri', $resource->uri ,80,200);

//  $out .= relationship_available_properties_form($subject);

  $terms = relationship_get_terms(); // If we don't know the TYPE of this resource, don't filter the available terms

  $out .= relationship_form($resource,$terms);

  $out .= form_button(t('Submit') );
  
  print( theme('page',form( $out ) ));

}



/**
 * A version of drupal form_hidden(),  but allowing me to 
 * set attribute on it too (the ID for css magic)
 */
function form_hidden_atts($name,  $value,  $attributes) { 
  return '<input type="hidden" name="edit['. $name .']" value="'. check_plain($value) ."\" ".drupal_attributes($attributes)."/>\n"; 
} 

/*! 
 *  @}
 */


?>
