<?php

/**
 * @file
 * Applies the logical inferences arising for the characteristics of predicates.
 * 
 * Characteristics are 'pluggable' as extensions. 
 * A new type of relationship may be defined by inserting the appropriate
 * definition into the database, as demonstrated below.
 *
 * If a characteristic supplies the appropriate interface,
 * (a function named as apply_characteristic_(characteristicName}() )
 * it may add deduced statements to the statement stack at recursion time.
 *
 * This file declares and impliments a useful set of characteristics. More can be
 * added by implimenting and declaring a characteristic in your own code.
 */
 
/**
 * @page characteristics About Characteristics
 *
 * Each apply_characteristic function hook is called by an inference resolver
 * with the following arguments:
 *
 * @parameter $ds datastore object (pass by ref).
 *            This serializes and sorts statements as they are discoverd.
 *            In indexing cache allows it to respond to queries once its
 *            been populated with facts. 
 *            @see rdf_datastore.
 *            It should be possible to just query these functions WITHOUT a 
 *            datastore object and just get back the resulting statements
 *            but inferences are much better when one is supplied.
 *
 * @parameter $subject string node ID , path, alias or URL
 *
 * @parameter $term_definition object,  optional - A term object stating the characteristics
 *            of the current predicate. Without it ALL statements with the subject as an 
 *            object will be returned
 *
 * @parameter $aliases optional array. A list of other names the subject may be 
 *            known as,  used for database lookups or create OR clauses.
 *
 * @return Array of new statements found.
 *         Modifications happen directly on the datastore object, but
 *         the new items should also be returned.
 *
 */


/**
 * Returns a list of definitions of characteristic terms currently supported
 */
function ontology_characteristics_defined_terms(){
  return array(
    PREDICATE_INVERSE_OF      => get_characteristic_inverseOf(),
    PREDICATE_SUB_PROPERTY_OF => get_characteristic_subPropertyOf(),
    'TransitiveProperty'      => get_characteristic_TransitiveProperty(),
    'UltimatePropertyOf'      => get_characteristic_UltimatePropertyOf(),
  );
}
 
$all_characteristics = relationship_get_all_characteristics();

/**
 * <rdf:Property rdf:ID="inverseOf">
 * <rdfs:label>inverseOf</rdfs:label> 
 * <rdfs:domain rdf:resource="#ObjectProperty" /> 
 * <rdfs:range rdf:resource="#ObjectProperty" /> 
 * </rdf:Property>
 */

function get_characteristic_inverseOf() {
  return array(
    'uri'          => OWL_NS. PREDICATE_INVERSE_OF,
    'name'         => PREDICATE_INVERSE_OF,
    'label'        => 'is the inverse (reciprocal relationship) of',
    'type'         => CLASS_PROPERTY,
    'description'  => '
        If A bears a relation to B,  then B implicitly has the reciprocal
        relation to B,  
        eg "IsFollowedBy" has the inverse relation "IsPrecededBy"
    ',
    'domain'       => CLASS_OBJECT_PROPERTY,
    'range'        => CLASS_OBJECT_PROPERTY,
    'isPartOf'     => ONTOLOGY_OWL,
    'sortOrder'    => 5,
  );
}

// Publish this logic routine
if (! $all_characteristics[PREDICATE_INVERSE_OF]  ) {
  relationship_import_definition( get_characteristic_inverseOf(), _relationship_characteristic_property_map() );
}

/**
 * Return all statements which imply the INVERSE of the given statement
 * or all statements which refer to the given subject.
 *
 * EG ,  if asked to apply_characteristic_inverse($ds, 'node1', 'IsParentOf')
 * The return would be the set of all statements claiming 
 * nodeX HasParent node1
 * - but rewritten into the inverse statement to place our node1 as the object again:
 * node1 IsParentOf nodeX (depth 1)
 *
 * If aliases is NULL,  check if the subject is known by any other names,  but 
 * If aliases is given as an empty array,  it's already been checked,  don't lookup again
 * This func impliments minor caching.
 *
 */
function apply_characteristic_inverseOf(&$ds,  $subject,  $term_def,  $aliases=NULL,  $depth=1,  $max_recursion_level=1) {

  if(is_array($subject)){
    $statement=$subject;
    $subject = $statement['subject'];
  }

  static $statement_cache; // avoid another DB lookup
  $new_statements = array();

  if ($term_def) {
    $reciprocal = $term_def->characteristics['inverseOf'];
    if (!$reciprocal) { 
      debug("Asked for inverse of the term '{$term_def->name}',  but it doesn't have one. Returning none.", 3);
      return; 
    }
    $predicate_clause = ' AND predicate = "'.$reciprocal.'" ';
  } // else no term,  get EVERY statement that has the subject as the object term. All incoming references

  debug("Scanning for inverse statements implying $subject,  ".($term_def ? $term_def->name : 'is referred to by')." ... anything. Aliases: ".join(', ',$aliases), 3);

  if (is_array($statements = $statement_cache[$subject][$reciprocal]) ) {
    debug("Returning cached inverse result ".print_r($statement_cache[$subject][$reciprocal], 1) , 3);
    if($ds){
      $ds->add_array(
        $statement_cache[$subject][$reciprocal], 
        $max_recursion_level - $depth, 
        'from inverseOf cache'
      );
    }
    return $statement_cache[$subject][$reciprocal];
  }

  // Make DB lookup
  
  $where_clause = ' object = "'.$subject.'" ';

  if (is_array($aliases) ) {
    // Need to search for nodes that may mention that also
    foreach ($aliases as $alias) {
      $where_clause .= ' OR object = "'.$alias.'" ';
    }
  }
  
  // Make a lookup to discover 
  // INFERRED relationship.

  $result = db_query('SELECT s.subject,  s.predicate,  s.object FROM {relationship_statement} s WHERE ('.$where_clause.') '.$predicate_clause.' ORDER BY predicate',  $subject);

  $statement_cache[$subject][$reciprocal] = array();
  while ($statement = db_fetch_array($result) ) {
    
    $statement['depth'] = $depth;

    debug("DB lookup Tells us "._relationship_hash_key($statement)." (when retrieving inverse statements about $subject)",2 );    

    if($ds){
      $ds->add(
        $statement, 
        $max_recursion_level - $depth, 
        "Database lookup when checking inverse relationships" 
      );
    }
    
    if(! $inverse = inverse_statement($statement) ){
      debug("Although this is an implication, there is no reciprocal relationship defined. Can't do anything with this back-reference",2 );    
      continue;
    };

    $inverse['depth'] = $depth+1;

    // remember for next time
    $key = _relationship_hash_key($inverse);
    $statement_cache[$subject][$reciprocal][$key]   = $inverse;

    debug("inverse lookup returns $key",2 );    

    if($ds){
      $ds->add(
        $statement, 
        $max_recursion_level - $depth, 
        "{$key['predicate']} is an inverseOf {$statement['predicate']}", $statement 
      );
    }

  } // while
  
  // the return is usually redundant, but informative
  return $statement_cache[$subject][$reciprocal];
}

/**
 * Reverse the operands and inverse the predicate for a statement.
 *
 * Eg turn a IsChildOf B into B HasChild A
 * If you already know the reciprocal,  pass it in,
 * otherwise it will be retrieved.
 *
 * Helper for apply_characteristic_inverseOf()
 *
 * @param $statement @see statements
 * @param $reciprocal predicate term definition @see predicates
 * @return modified statement,  also changes it by reference. 
 *         Returns null if no inverse predicate exists
 */
function inverse_statement(&$statement, $reciprocal='') {
  if (! $reciprocal ) {
    $term_def = relationship_get_term($statement['predicate']);
    if(!$term_def){return;} // no such relationship
    $reciprocal = $term_def->characteristics['inverseOf'];
  }
    
  if ($reciprocal) {
    $recip_statement = array(
      'subject'   => $statement['object'],
      'predicate' => $reciprocal,
      'object'    => $statement['subject'],  
      'inverseOf' => $statement['predicate'],  
      'title'     => '',
      'depth'     => 1
   );
  }
  return $statement = $recip_statement;
}

 






/**
 * subPropertyOf is a _characteristic_ of a type of relationship.
 *
 * isSonOf is a subPropertyOf IsChildOf - it is a refinement of the
 * original statement,  and implies that when A is subPropertyOf B,  then
 * A(x, y) implies B(x, y) also.
 * In practice that mean that any statement using predicate A is 
 * re-iterated with predicate B. This function ADDS new statements to the
 * $statement_stack,  and returns how many new statements were created.
 *
 * According to 
 * http://www.w3.org/TR/rdf-schema/
 *
 * <rdf:Property rdf:about="http://www.w3.org/2000/01/rdf-schema#subPropertyOf">
 *  <rdfs:isDefinedBy rdf:resource="http://www.w3.org/2000/01/rdf-schema#" /> 
 *  <rdfs:label>subPropertyOf</rdfs:label> 
 *  <rdfs:comment>The subject is a subproperty of a property.</rdfs:comment> 
 *  <rdfs:range rdf:resource="http://www.w3.org/1999/02/22-rdf-syntax-ns#Property" /> 
 *  <rdfs:domain rdf:resource="http://www.w3.org/1999/02/22-rdf-syntax-ns#Property" /> 
 * </rdf:Property>
 *
 */
function get_characteristic_subPropertyOf() {
  return array(
    'uri'          => RDFS_NS . PREDICATE_SUB_PROPERTY_OF,
    'name'         => PREDICATE_SUB_PROPERTY_OF,
    'label'        => 'is a subproperty (specific type) of',
    'type'         => CLASS_PROPERTY,
    'description'  => '
        A refinement of the original statement.
        isSonOf is a subPropertyOf IsChildOf - it implies that when 
        A is subPropertyOf B,  
        then A(x, y) implies B(x, y) also.
        Almost all relationship are really refinements of "IsRelatedTo". 
        
        "IsDaughterOf" would be a refinement to "IsChildOf", so when array 
        walking is done to retrieve all children of an item, the daughters 
        may be returned also. 
        
        "IsChapterOf" may be a subPropertyOf "IsisPartOf", so retrieving all 
        subsections (non-transitive) would be more specific (and useful) 
        than retrieving all "parts", which would include appendices, 
        forwords etc.
    ',
   'domain'       => CLASS_OBJECT_PROPERTY,
   'range'        => CLASS_OBJECT_PROPERTY,
   'isPartOf'       => ONTOLOGY_OWL,
     
  );
}

/* Insert this definition into the database */
 
if (! $all_characteristics[PREDICATE_SUB_PROPERTY_OF] ) {
  relationship_import_definition( get_characteristic_subPropertyOf(), _relationship_characteristic_property_map() );
}

 
function apply_characteristic_subPropertyOf(&$ds,  $subject,  $term_def,  $aliases=NULL,  $depth=1,  $max_recursion_level=1) {

  if(is_array($subject)){
    $statement=$subject;
    $subject = $statement['subject'];
  }

  debug("Deducing additional statements as the {$term_def->name} relationship is a subPropertyOf {$term_def->characteristics['subPropertyOf']}", 3);
  $super_term = $term_def->characteristics['subPropertyOf'];
  $new_statements = array();

  if ($statement['predicate'] == $term_def->name)
  {
    // for each statement made using the sub-term,  the same is implicitly true
    // for the super-term.
    // Apply it as a new statement
    $new_statement = array(
        'subject'  => $subject,
        'predicate'=> $super_term,
        'object'   =>$statement['object'],
        'depth'    => $depth
    );

    $ds->add($new_statement, $max_recursion_level - $depth, " {$term_def->name} is a subPropertyOf $super_term", $statement);

    $new_statements[] = $new_statement;
  }
  return $new_statements;
}








/**
 * TransitiveProperty is a _characteristic_ of a relationship property.
 *
 * If relationship A is transitive,  then A(x, y) and A(y, z) implies A(x, z)
 * Applying this property to a statement stack will add new statements
 * that state this new relationship explicitly.
 *
 * For all statements made so far,  IF the stated relationship is transitive
 * AND the object of the statement also has one of these relationship,
 * a new implied statement linking the current subject with the objects object
 * is declared.
 *
 * http://www.w3.org/TR/2004/REC-owl-ref-20040210/#TransitiveProperty-def
 * Problem - 'transitive' as defined by OWL is a Class, not a property.
 * I can't handle that properly yet.
 *
 * <rdfs:Class rdf:about="http://www.w3.org/2000/01/rdf-schema#TransitiveProperty">
 *  <rdfs:label>TransitiveProperty</rdfs:label> 
 *  <rdfs:subClassOf rdf:resource="#ObjectProperty" /> 
 * </rdfs:Class>
 *
 */
function get_characteristic_TransitiveProperty() {
  return array(
    'name'         => 'TransitiveProperty',
    'label'        => 'TransitiveProperty',
    'uri'          => 'http://www.w3.org/2000/01/rdf-schema#TransitiveProperty',
    'subClassOf'   => CLASS_OBJECT_PROPERTY,
    'description'  => '
        If A=>B and B=>C then A=>C , eg "isAncestorOf", "isPartOf"
    ',
   'domain'       => CLASS_OBJECT_PROPERTY,
   'range'        => CLASS_LITERAL,
   'isPartOf'       => ONTOLOGY_OWL,
   );
}
// Publish this logic routine

if (! $all_characteristics['TransitiveProperty'] ) {
  relationship_import_definition( get_characteristic_TransitiveProperty(), _relationship_characteristic_property_map() );
}
 
/**
 * Implimentation of apply_characteristic_hook() 
 */
function apply_characteristic_TransitiveProperty(&$ds,  $subject,  $term_def,  $aliases=NULL,  $depth=1,  $max_recursion_level=2) {

  if(is_array($subject)){
    $statement=$subject;
    $subject = $statement['subject'];
  }

  debug("Deducing additional statements as the {$term_def->name} relationship is transitive", 3);
  $new_statements = array(); 

    if ($statement['predicate'] == $term_def->name)
    {
      // For each statement made using the transitive term,  
      // see if the target also has a relationship like that.
      // If so,  link this subject with each of the objects objects.
      if (! $statement['object']) {continue;}

      debug("Looking up to see if {$statement['object']} is also {$term_def->name} anything ", 3);

//print("<fieldset><legend>{$statement['object']} {$term_def->name} lookup $depth</legend>");
      
      // to recurse or not?
      $sub_statements = relationship_query($statement['object'],  $term_def->name );

//print("</fieldset>");

      if (is_array($sub_statements) ) {

      debug("Lookup returned ".implode(", ",array_keys($sub_statements)), 3);

        foreach ($sub_statements as $sub_statement) {
          // its possible for get_statement lookups to return too much information - 
          // Relationships they deduced on the way. We can filter that out again.
          
          // .. or does datastore do the filtering now ?
          if ($sub_statement['predicate'] == $term_def->name) {
            $new_statement = array(
              'subject'  => $subject,
              'predicate'=> $term_def->name,
              'object'   => $sub_statement['object'],
              'depth'    => $depth
            );
           
            debug("Found implied relationship. Not only is $subject {$term_def->name} {$statement['object']},  but {$sub_statement['subject']} {$sub_statement['predicate']} {$sub_statement['object']} . hence $subject {$term_def->name} {$sub_statement['object']}  ", 2);

            $ds->add($new_statement, $max_recursion_level - $depth, "{$term_def->name} is a transitiveProperty",$sub_statement);
            $new_statements[] = $new_statement;
          } // if the right predicate in results
        }// loop all lookup results
      }// if any results
      else {
        debug("No results in transitiveProperty ",2);
      }
    }// if the right predicate in source

  return $new_statements;
}





/**
 * UltimateProperty is a _characteristic_ of a relationship property.
 * It represents the last relationship found when following a linked
 * chain of relationships.
 * The property "last" would be the UltimateProperty of the "next" relationship,
 * The property "first" would be the UltimateProperty of the "previous" relationship,
 * The property "LatestVersion" would be the UltimateProperty of the "replacedBy" relationship.
 *
 * It's like 'inheiritedFrom' but doesn't return the first answer it finds,  but the last
 * A relationship defined as an UltimateProperty should not normally be stated explicitly,  
 * but always inferred.
 * In the unlikely case of the lookups branching,  multiple UltimateProperty
 * may be returned,  although this often indicated a structural anomaly elsewhere.
 *
 * It can be imagined that a document IsReplacedBy two or more independant specifications,
 * (eg the modularization of XHTML http://www.w3.org/MarkUp/modularization)
 * This would imply that the LatestVersionOf the old document would be the 
 * LatestVersionOf BOTH replacements.
 *
 * As this statement is seldom explicitly stated,  it may be called with an undefined subject.
 * Just the currently defined statements are scanned to see if the dependant (linked) property is 
 * available. If so,  logic is deduced from there.
 */
 
function get_characteristic_UltimatePropertyOf() {
  return array(
    'name'         => 'UltimatePropertyOf',
    'label'        => 'is the last value in the chain of!',
    'type'         => CLASS_PROPERTY,
    'description'  => "
         It represents the last relationship found when following a linked
         chain of relationships.
         The property 'last' would be the UltimateProperty of the 'next' relationship,
         The property 'first' would be the UltimateProperty of the 'previous' relationship,
         The property 'LatestVersion' would be the UltimateProperty of the 'replacedBy' relationship.

       It's like 'inheiritedFrom' but doesn't return the first answer it finds,  but the last.
    ",
   'domain'       => CLASS_OBJECT_PROPERTY,
   'range'        => CLASS_OBJECT_PROPERTY,
   'isPartOf'       => ONTOLOGY_OWL,
     
  );
}
// Publish this logic routine
if (! $all_characteristics['UltimatePropertyOf'] ) {
  relationship_import_definition( get_characteristic_UltimatePropertyOf(), _relationship_characteristic_property_map() );
}

/**
 * Find the last relationship in a chain.
 *
 * Implimentation of apply_characteristic_hook()
 * @todo review.
 * should impliment 'cascades' first, and use that under this
 */
function apply_characteristic_UltimatePropertyOf(&$ds,  $subject,  $term_def,  $aliases=NULL,  $depth=1,  $max_recursion_level=2) {

  if(is_array($subject)){
    $statement=$subject;
    $subject = $statement['subject'];
  }

  $term = $term_def->name;
  $target_term = $term_def->characteristics['UltimatePropertyOf'];

  debug("
      Deducing additional statements as the '$term' relationship is defined as 
      being the UltimateProperty of '$target_term' 
      (depth $depth/$max_recursion_level) 
  ", 2);


  $results = array();
  debug("Looking up '$term' from '$subject' at depth $depth/$max_recursion_level", 3);

  $sub_relations = relationship_query($subject,  $target_term);

  if (is_array($sub_relations) ) {
    //    if (count($sub_relations) > 1) {debug("Found more than one '$target_term' from $subject when fetching statements. I can branch,  but you may not get the results you expect. Apparently ".implode(",  ", array_keys($sub_relations) ), 0);}
    foreach ($sub_relations as $sub_statement) {
      if ($sub_statement['subject']!=$subject) {continue;}

      debug("found a lookup OK $subject:$target_term:{$statement['object']}", 3);
      $key = _relationship_hash_key($sub_statement);
      
      if (! $ds->get_by_key($key) ) {
        $ds->add( $sub_statement, $max_recursion_level - $depth, 'UltimateProperty retrieval',$statement );

        $lookups = apply_characteristic_UltimatePropertyOf(
          $ds,  $sub_statement['object'], $term_def,  
          $aliases,  $depth+1,  $max_recursion_level
        );

        if (count($lookups) > 1) {
            debug("Found more than one '$target_term' from '$subject' when doing 
            UltimatePropertyOf(). I can branch,  but you may not get the results you expect 
        ", 0);}

        // Lookups return statements about higher - level relationships.
        // Make those relationships our own
        foreach ($lookups as $super_statement) {
          $super_statement['subject'] = $subject;
          $super_statement['predicate'] = $term;
          $super_statement['depth']++;
          #$key = _relationship_hash_key($super_statement);
          #$results[$key] = $super_statement;
          $new_statements[] = $super_statement;
          $ds->add( $statement, $max_recursion_level - $depth, 'UltimateProperty recursion', $sub_statement );
        }
        // each intermediate statement is also true.

        if (! $lookups) {
          $new_statement = array(
            'subject' => $subject,
            'predicate' => $term,
            'object' => $sub_statement['object'],
            'depth' => 1,
         );
         
          // return the topmost relationship
          $new_statements[] = $new_statement;
          $ds->add( $new_statement, $max_recursion_level - $depth, 'As the UltimateProperty',$sub_statement );
#          $key = _relationship_hash_key($statement);
#          $results[$key] = $statement; 
        }
      } else {
        debug("Been here before! Beware of looping. I've just been told $key .. again. Stopping now,  it may not be correct,  but what do you expect? the trail is:".implode(" - ", array_keys($statement_stack) ), 3);
        return $results;
      }
    }
  }
  
  debug("When looking up '$term' at depth $depth,   returning ".print_r($results, 1), 2);
  return $results;

}





/**
 * CascadeDeletion is a _characteristic_ of a relationship property.
 *
 * It indicates that this property should destroy all its children if the
 * parent is destroyed. An example would be a threaded discussion.
 * If the root node (or indeed anywhere in the chain) was intentionally deleted,
 * all comments below there have nowhere to live,  and should be terminated 
 * with extreme prejudice.
 * Be very careful to only apply this to twig,  'ChildOf' sort of relationship.
 * Implimentation of apply_characteristic_hook()
 */
function apply_characteristic_CascadeDeletion(&$statement_stack,  $subject,  $term_def,  $aliases=NULL,  $depth=1,  $max_recursion_level=2) {

  // This property doesn't actually do anything during the 
  // apply_characteristic logic stage.
  // It should kick in as a nodeAPI hook.
  
}


/*
   $glossary.='<dt><a href="http://www.w3.org/TR/2004/REC-owl-ref-20040210/#SymmetricProperty-def" title="See the W3C definition of this characteristic">Symmetric</a></dt><dd>If A=>B then B=>A. If A bears a relation to B,  then B implicitly has the <em>same</em> relation to B,  eg "IsNeighbourOf" "IsSiblingOf".  No second argument.</dd>';
  $glossary.='<dt><a href="http://www.w3.org/TR/2004/REC-owl-ref-20040210/#TransitiveProperty-def" title="See the W3C definition of this characteristic">Transitive</a></dt><dd>If  A=>B and B=>C then A=>C ,  eg "IsAncestorOf",  "IsisPartOf". No second argument.</dd>';
  $glossary.='<dt><a href="http://www.w3.org/TR/2004/REC-owl-ref-20040210/#subPropertyOf-def" title="See the W3C definition of this characteristic">Subproperty</a></dt><dd>A refinement on a more general term. Almost all relationship are really refinements of "IsRelatedTo". "IsDaughterOf" would be a refinement to "IsChildOf",  so when array walking is done to retrieve all children of an item,  the daughters may be returned also. "IsChapterOf" may be a subPropertyOf "IsisPartOf",  so retrieving all subsections (non-transitive) would be more specific (and useful) than retrieving all "parts",  which would include appendices,  forwards etc.</dd>';
  $glossary.='<dt>Inheirited</dt><dd>If the property is <em>not</em> explicitly set for a resource,  retrieve property of the nearest other resource that does define it. eg if "Creator" is inheirited from "IsisPartOf" then the default "Creator" of a resource is the creator of whatever the resource "IsisPartOf". As IsisPartOf is transitive,  the tree may be walked until a "Creator" is found. Useful to define the contactee details for a page.</dd>';
  $glossary.='<dt>Cascaded</dt><dd>(cumulatively inheirited). Even if this property is set,  the linked relationship is walked and <em>all</em> defined relationship of this type are returned. An example may be "IsMaintainedBy" cascades from "IsisPartOf" so that a resource can be maintained  by <b>all</b> maintainers that maintain whatever subsections and sections this resource IsisPartOf. Useful for managing access rights or to send change notifications.</dd>';
<p>
      While isPartOf may usually seem the most natural property to inheirit from,
      alternatives are possible. "HasHairColour" may inheirit from "HasMother"
      while "HasSurname" may inheirit from "HasFather"
    </p>';

*/

?>