
/**
 * @file combobox.js
 * 
 * Support functions for a form administration interface
 * (Drupal)
 *
 * Support a DHTML combo box - allow either selectbox or freetext data entry in one field
 *
 * Support dynamic list management, can put an 'Add Row' or 'Delete Row'
 * button on a table to manage lists of fields.
 *
 * DHTML created with reference to that used in Drupal 'Menu On The Fly' module
 *
 * @author dan [at] coders.co.nz
 * @version Nov 2005
 */


// The class ID used to ID which select boxes may become freetext also
var COMBO_CLASS = "combo";



/**
 * Update all labelled combo boxes with a button that will change it to a text field instead.
 * The button is created programatically after page load, so should not effect anything with
 * noscript
 */
function init_combo_box(){
  if (document.getElementsByTagName) {
    var f = document.getElementsByTagName('select');
    for (var i=0; i<f.length; i++) {
      if ( (f[i].getAttribute('class') == COMBO_CLASS) ||
           (f[i].getAttribute('classname') == COMBO_CLASS) )
      {
        field_id = f[i].getAttribute('id')
        if(! field_id ){ 
          field_id = f[i].getAttribute('name');
          f[i].setAttribute('id',field_id); 
        }

        container = f[i].parentNode;
        container.setAttribute("id","combo_container_"+field_id);
        var newButton = null;
        
        // Look for a button already there, don't add if it's been done
        existing = container.getElementsByTagName("input");
        for(var ix in existing){
          if(existing[ix].getAttribute){
            if(existing[ix].getAttribute("type") == "button")
              newButton=existing[ix]
          }
        }

        if (! newButton ){
          newButton = document.createElement("input");
          newButton.setAttribute("type","button");
          newButton.setAttribute("value","T");
          newButton.setAttribute("title","TEXT: click here to enter a freetext URL or identifier");
          newButton.setAttribute("id","combo_button_"+field_id);
          container.appendChild(newButton); 
        }
        // Do add the event each time, as DOM manipulation doesn't do that for us.
        addEvent(newButton, 'click', switch_to_freetext);

      }//if
    }//for
  }// if
}

/**
 * Change the type of input from Select to Text.
 * Called when the button is pressed. Actually discards the select
 * (so you can't get it back) and makes a new field with the same name
 */
function switch_to_freetext(event){
  if(! (e = event.srcElement)){ e = event.target }
  if(! e ) {return;}

  var id = e.id.substring(13);
  
  var element = document.getElementById(id);
  if(element){
    change_element_to_freetext(element)
  } else {
    alert("Failed to find element "+id)
  }
}

function change_element_to_freetext(element){

  var newField = document.createElement("input");
  newField.setAttribute("type","text");
  newField.setAttribute("name", element.getAttribute("name"));
  newField.setAttribute("size", 40);

  if(element.options && (element.selectedIndex >= 0) ){
    newField.setAttribute("value", element.options[element.selectedIndex].value);
  }
  else
    newField.setAttribute("value", element.getAttribute("value"));
    
  element.parentNode.insertBefore(newField,element);
  id = element.id;
  element.parentNode.removeChild(element);
  newField.id = id;
  
}

function change_element_to_select(element,indexed_options){
  var newField = document.createElement("select");
  newField.setAttribute("name", element.getAttribute("name"));

  currentval=element.getAttribute('value');
  if(!currentval){currentval = element.options[element.selectedIndex].value;}

  for(o in indexed_options){
    add_option_to_select(newField, o, indexed_options[o],currentval );
  }

  element.parentNode.insertBefore(newField,element);
  id = element.id;
  element.parentNode.removeChild(element);
  newField.id = id;
  
}

// allow recusive option groups
// IE doesn't display nested optgroups :(
function add_option_to_select(select,label,list,selected){

  if(typeof(list) == 'object'){
    // the options are grouped
    og = document.createElement("optgroup");
    og.setAttribute("label",label);
    select.appendChild(og);

    for(varname in list)
    {
      add_option_to_select(og, varname, list[varname],selected);
    }
    
  } else {
    op = document.createElement("option");
    op.setAttribute("value",label);
    op.appendChild(document.createTextNode(list));
    select.appendChild(op);
    if(label == selected){op.setAttribute('selected','selected')} 
  }    
}

addEvent(window, 'load', init_combo_box);



/**
 * Add a new row to a table by duplicating the last row.
 * This allows users to keep adding new info to a table that includes 
 * editable text fields as the last row.
 * Should be triggered from a button of your own that calls this function.
 * Table must have an 'id' set
 */
function add_row_to_edit_table(table_id){
  if (document.getElementById) {

    var t = document.getElementById(table_id);
    if(t){
      // find last row, 
      var rows = t.getElementsByTagName("tr");
      var lastrow = rows[rows.length-1];
      var copyrow = lastrow.cloneNode(true);
      // Cannot allow duplicate IDs
      // By changing the old row, hopefully scripts will still work on the new row
      _set_unique_ids(lastrow,rows.length)

      lastrow.parentNode.appendChild(copyrow);

      // need to do this to add the event to the button again
      init_combo_box();
    } else {
      alert(table_id+" not found, cannot add a row to it");
    }
    return false;
  }
  return true
}

/**
 * Recursively ensure that any item in this tree with an id gets suffix added to it
 */
function _set_unique_ids(node,suffix){
  var child = node.firstChild;
  if(!child) return;
  while (child){
    if(child.id){
      child.id = child.id + suffix ;
    }
    _set_unique_ids(child,suffix);
    child = child.nextSibling;
  }
}




/**
 * Remove the target row from a table.
 * Called from a button with the parameter 'this', this func
 * will find the nearest tr and wipe it.
 */
function delete_row_from_edit_table(event_src){
  if (document.getElementById) {
    parentN = event_src['parentNode']; // 'parent' is a reserved word in IE!

    while(parentN && (parentN != parentN.parentNode) && (parentN.nodeName.toLowerCase() != "tr") ){
      parentN = parentN.parentNode;
    }
    if(parentN.nodeName.toLowerCase() == "tr"){
      // and remove it
      parentN.parentNode.removeChild(parentN);
    } else {
      alert("No row found above here - was this action launched from inside a table at all?");
    }
    return false;
  }
  return true;
}


/**
 * example :
 * <div id='changeable_button_foo'>
 * <input class='changeable' type='hidden' value='myval' name='myname' id='hidden_foo' />
 * <span id='exposed_foo'>My name is 'myname'</span>
 * </div>
 *
 * The input class:changeable will be found, and the id 'foo' deduced from it.
 * an onClick event is added to the div/span/whatever.
 * When clicked, the hidden input will be removed and the exposed span will be replaced
 * with an equivalent textfield.
 * @author .dan. dan[at]coders.co.nz
 */

// The class ID used to ID which hidden fields may become freetext also
var CHANGEABLE_CLASS = "changeable";
var CHANGEABLE_BUTTON_PREFIX = "changeable_button_"; // added to the button id
var CHANGEABLE_HIDDEN_PREFIX = "hidden_" // added to the hidden field id
var CHANGEABLE_EXPOSED_PREFIX = "exposed_" // added to the hidden field id

/**
 * Show a previously 'hidden' field in the page
 * A Hidden input will be replaced dynamically with a textfield
 * inline to its original location. 
 * If an optional second argument
 * is given (the ID to the previous, uneditable label) that element
 * is removed from the page.
 * If called as an event handler (one paramenter - the event)
 * then it tries to find the appropriate ids from context.
 */
 function expose_hidden_field(id, del_id){
  if (document.getElementById) {

    if(!del_id){ // called as an event handler, ok figure out our context
      if(! (e = id.srcElement)){ e = id.target}
      if(! e ) {return;}
      // may need to recurse up - the event may have been too specific

      while(e.parentNode && (e.parentNode != e) && (! e.id.match(CHANGEABLE_BUTTON_PREFIX) )){
        e = e.parentNode;
      }


      id = e.id.substring(CHANGEABLE_BUTTON_PREFIX.length);
      del_id = CHANGEABLE_EXPOSED_PREFIX + id.substring(CHANGEABLE_HIDDEN_PREFIX.length)
    }
    
    var f = document.getElementById(id);
    if(f){
      container = f.parentNode;
      var newField = document.createElement("input");
      newField.setAttribute("type","text");
      newField.setAttribute("name",f.getAttribute("name"));
      newField.setAttribute("value",f.getAttribute("value"));
      newField.setAttribute("size", 50);
      container.insertBefore(newField,f); 
      container.removeChild(f);
      
      // remove the older label that may have been a placeholder
      var old = document.getElementById(del_id);
      if(old){
        old.parentNode.removeChild(old);
      }
      var old = document.getElementById(CHANGEABLE_BUTTON_PREFIX+id);
      if(old){
       removeEvent( old, 'click' , expose_hidden_field );
      }
     
      // need to cancel the event now.
      return 1 ; 
      
    } else alert("Failed to find "+id)
  } 
}

/**
 * Update all labelled 'changeable' hidden fields so they can be exposed with a click.
 * The button is created programatically after page load, so should not effect anything with
 * noscript
 */
function init_changeable_fields(){
  if (document.getElementsByTagName) {
    var f = document.getElementsByTagName('input');
    for (var i=0; i<f.length; i++) {
      if ( (f[i].getAttribute('class') == CHANGEABLE_CLASS) ||
           (f[i].getAttribute('classname') == CHANGEABLE_CLASS) )
      {
        field_id = f[i].getAttribute('id')
        if(! field_id ){ 
          field_id = f[i].getAttribute('name');
          f[i].setAttribute('id',field_id); 
        }

        container = f[i].parentNode;
        container.setAttribute("id",CHANGEABLE_BUTTON_PREFIX+field_id);
        addEvent(container, 'click', expose_hidden_field);

      }//if
    }//for
  }// if
}
addEvent(window, 'load', init_changeable_fields);




// Utils.
// credits: http://ejohn.org/projects/flexible-javascript-events/
function addEvent( obj, type, fn ) { 
  if ( obj.attachEvent ) { 
    obj['e'+type+fn] = fn; 
    obj[type+fn] = function(){obj['e'+type+fn]( window.event );} 
    obj.attachEvent( 'on'+type, obj[type+fn] ); 
  } else 
    if(typeof('obj.addEventListener') == 'function')
      obj.addEventListener( type, fn, false ); 
} 
function removeEvent( obj, type, fn )
{
  if (obj.removeEventListener)
    obj.removeEventListener( type, fn, false );
  else if (obj.detachEvent)
  {
    obj.detachEvent( "on"+type, obj[type+fn] );
    obj[type+fn] = null;
    obj["e"+type+fn] = null;
  }
}




    function dump(o){
      $out ="";
      for(p in o){
        $out += "\t"+p+" = " + o[p] + "("+typeof(o[p])+")"; 
      }
      return $out;
    }
