<?php
/**
 * @file
 * Library to extend the 'object' selector to show only appropriate targets
 * If the 'author' property should only point to users then the 'range' of
 * 'author' is 'users' and someone attempting to set the 'author' should 
 * be presented with a list of users.
 * Today we'll do it by populating a select via javascript.
 * Options are to AJAX something in the background ... next
 */

/**
 * Returns javascript to enable the dynamic selectbox updates
 */
function range_selectors($predicate_list, $trigger_prefix, $target_prefix) {

  if(! is_array($predicate_list)){
    debug("Invalid arg '$predicate_list' to range_selectors() ");
    return;
  }

  $relevant_ranges = relationship_get_selector_ranges($predicate_list);

  // ensure that 'nodes' is a valid range that is always available
  $relevant_ranges['default'] = array('node','user','vocabulary');
  $relevant_ranges[CLASS_RESOURCE] = array_unique ( array_merge($relevant_ranges[CLASS_RESOURCE], relationship_get_range_names() ) );

  // Create a javascript config array for this info 
  // ... may go on-demand later, but inline is easier
  // This defines the mapping between predicates and ranges
  
  $js = "\n// Listing which selection values (on the left) should populate which ranges (on the right)\n";
  $js .= "// Each selection may refer to more than one range, so it's initialized with an array of range names,\n";

  $js .= js_array_dump($relevant_ranges ,'selector_ranges' );

  // Now serialize the range values themselves
  $required_range_definitions = array();
  foreach($relevant_ranges as $predicate => $ranges){
    if( is_array($ranges) ){
      foreach($ranges as $range_name){
        if(! $required_range_definitions[$range_name] ){
          $required_range_definitions[$range_name] = relationship_get_range($range_name);
        }
      }
    }
  }
  
  $js .= "\n\n// Now define the contents of the ranges\n";
  $js .= js_array_dump($required_range_definitions ,'ranges' );


  $js .= "\nvar TRIGGER_PREFIX = '$trigger_prefix'; var TARGET_PREFIX = '$target_prefix';\n";

//  $js .= "\n\n// A property with a domain set as the superclass 'Resource' can be any URI\n";
//  $js .= "ranges['Resource'] = ranges.slice(0)// (a copy)\n"; // loop?

  $js = "<script>\n$js\n</script>";

  return $js;
}



/**
 * Return the onchange event that will cause a change in the trigger selector
 * to update the target selector.
 */
function range_selectors_trigger_update($trigger,$target) {
  return " onchange=' change_select_range(this); ' ";
}



/**
 * Return a list of all defined content types.
 *
 * ... plus a few built-ins like 'users' and 'vocabularies'
 * Each of these define ranges - possible sets of objects.
 */
function relationship_get_range_names() {
  $range_names = array();
  $range_names = array_map( create_function('&$n', 'return "node/".$n;'), node_list() ) ;
  array_push($range_names,'user');
  $vocabs = array_map('relationship_get_shortname_property',taxonomy_get_vocabularies());
  $vocab_names = array_map( create_function('&$n', 'return "vocabulary/".$n;'), $vocabs ) ;
  return array_merge($range_names,$vocab_names);  
}
 
 
 /**
  * Return the values in a given range, as named in relationship_get_range_names().
  *
  * A range name may be 'node/nodetype' or 'vocabulary/vocabname'
  * Internal ranges are all predicates of a given 'type', like
  * 'property', 'Class', or 'Ontology'.
  * more?
  */
function relationship_get_range($range_name) {
  $range = array();
  $bits = explode('/',$range_name);
  switch ($bits[0]) {
    case CLASS_LITERAL :
      // No range at all. 
      break;
    case 'node' :
      if($bits[1]){
        $where = " WHERE type = '%s' ";
      }
      // isn't there an API for this request?
      // copied from the snippets
      $sql = "SELECT title, nid FROM {node} $where ";
      $result = db_query($sql,$bits[1]);

      while ($anode = db_fetch_object($result)) {
        $range[$anode->nid] = $anode->title;
      }
    break;
    
    case 'vocabulary' :
      $vocabs = array_flip(array_map('relationship_get_shortname_property',taxonomy_get_vocabularies()) );

      if($vocab = $bits[1]){
        // if defined, return just terms from given vocab, else return terms from all
        $vocabs = array($vocab => $vocabs[$vocab] );
      }

      foreach($vocabs as $vocab => $vid ){
        $tree = taxonomy_get_tree($vid);

        // from _taxonomy_term_select()   
        if ($tree) {
          foreach ($tree as $term) {
            $range['term/'.$term->tid] = $term->name;
          }
        }
      }
      break;
      
    case 'user' :
      // no API for user list either? lifted from user.module
      // Users are identified by the path to their user profile
      $sql = 'SELECT uid, name FROM {users}  ';
      $result = db_query($sql);
      while ($auser = db_fetch_object($result)) {
        if($auser->name){
          $range['user/'.$auser->uid] = $auser->name;
        }
      }
      break;
    case 'types' :
      $range = relationship_get_range_names();
      break;
    default :
      // try to retrieve everything that is of the given 'type';
      debug("Getting range ".$bits[0]." - searching for all resources of type '".$bits[0]."'" ,3 );
      $range =  array_flip( relationship_retrieve_subjects_by_predicate(PREDICATE_TYPE,$bits[0]) );
      foreach($range as $id => $ix){
        $range[$id] = relationship_get_shortname($id);
      }
   
  } // switch
  
  if($bits[0] == 'Class'){
    // as well as everything literally called a class,
    // Add all locally defined contact types as valid 'classes'
    foreach (node_list() as $type){
      $range[$type] = node_invoke($type, 'node_name');
    }
    // And out local vocabularies
    foreach (taxonomy_get_vocabularies() as $vocabulary){
      $range['taxonomy/term/'.$vocabulary->vid] = $vocabulary->name ;
    }
    
  }
 
 if(! $range && ($range_name != CLASS_LITERAL)){ 
   debug("no results for range '$range_name' when populating the array for the selectbox. Basically this means there is no possible valid value for this predicate. Odd.",2);
 }
 
 return $range;
}

/**
 * Util for array_map on objects
 */
function relationship_get_shortname_property($target){return $target->name;}

/**
 * lookup a list of predicates and return a list of the union of their respective ranges.
 *
 * The list can be indexed on uri or shortname.
 * The content of the list can be redundant OR a handle on the predicate
 * definition object itself. If given, this will save a lookup.
 *
 * @todo deal with shortnames - currently
 * indexing on URIs instead
 */
function relationship_get_selector_ranges($predicates) {
  $ranges = array();
  
  if(! is_array($predicates)){
    trigger_error("Invalid arg '$predicates' to relationship_get_selector_ranges() ",E_USER_WARNING);
    return array();
  }
  
  $relationship_long_names = variable_get('relationship_long_names', FALSE);

  foreach($predicates as $predicate => $def){
    // index on shortname or longname? 
    $indexid = ($relationship_long_names)?$def->uri:$def->name;

    if(is_object($def)){
      if($def->range){
        // ranges may be multiple. assume the range into an arrays
        if(! is_array($def->range)){ $def->range = array( $def->range ); }
        $ranges[$indexid] = $def->range;  
        $ranges[CLASS_RESOURCE] = array_merge($ranges[CLASS_RESOURCE],$ranges[$predicate]);
      }
    } else {
      $ranges[$indexid] = relationship_get_properties($predicate,  PREDICATE_RANGE);
    }

    debug("The range of $predicate $label is ".print_r($ranges[$indexid],1),2 );
  }
  return $ranges;
}

/**
 * Renders an array in a javascript way.
 *
 * This should be put into a script block, even better as an include.
 * @todo figure out caching for this.
 */
function js_array_dump($ar,$name = NULL) {

  if(is_array($ar)){
    $name = preg_replace('/[^\w]/','_',$name);
    if( $name ) {$out .= "\n\nvar $name = " ;}
    else {$padd='  ';}

    $out .= "$padd{\n";
    $vals = array();
    foreach($ar as $var => $val){
      if(is_array($val)){
        $vals[] = "'$var':". js_array_dump($val) ."\n" ;
      } else {
        $vals[] = "  '$var':'".check_plain($val)."' ";
      }
    }
    $out .= join(",$padd\n",$vals);
    $out .= "$padd\n}";
  }
  
  return $out;
  
}

?>
