<?php

/*
 * Dear coder
 *
 * If you are looking at this code, It's probably best to start by having a look
 * at docs.php or the autogenerated API docs. Theres a conceptual overview over there.
 * .dan.
 */

/**
 * @file relationship api
 * @author Dan Morrison <dan [AT] coders.co.nz>
 * 
 * Interface for modules and other code to set and get relationship statements.
 *
 * This layer takes care of the datastore abstractions, and
 * provides a set of utility accessor functions to establish and query
 * relationships between resources, defined as subject-predicate-object statements.
 */

/**
 * @ingroup relationship_api
 * @{
 */

/*
 * Statement cache object
 * @todo make this more pluggable
 * The idea is that different datastores may use different databses 
 * (or even server queries) in different ways.
 * Thus, all my 'set' and 'queries' are interpreted by a black-box.
 */
include_once('rdf_datastore.inc');


/////////////////////////////////////////////////////////
// Programmers API
/////////////////////////////////////////////////////////
// Accessor Functions
//
//  Set,  get,  lookup terms and relations
//


///////////////////////////////////////
// Constructors & converters


/**
 * Just a helpful struct array constructor
 *
 * About the 'statement' model used internally:
 *
 * To retrieve the metadata for a node:
 *
 *     $statements = relationship_query($nid);
 *
 * Returns an array of statements.
 * Each statement may look something like:
 *
 *     $statement = Array 
 *         (
 *             [subject] => 198            // The current node
 *             [predicate] => isPartOf     // The relationship identifier
 *             [object] => 197             // The target node OR ALIAS
 *             [title] => Chapter 1        // The display title,  extracted from the target node itself
 *             [object_nid] => 197         // The target node ID,  useful if the defined 'object' was an alias path
 *        )
 *
 * Optionally,  it may also have a [depth] value set,  which indicates
 * how much logical recursion it took to come up with this statement.
 * A depth > 0 indicated this statement is IMPLIED,  not explicitly stated.
 * A depth < 0 is trivial, used for system statements
 * like a Drupal Alias is interpreted as a RDF:sameAs relationship
 * and other Drupal-managed data, like lastModifiedDate which may sometimes
 * be represented as statements, but is not expected to be serialized as such.
 *
 * Another optional property is
 *             [path] => node/197          // The link to display to this target
 * this gets calculated if needed when relationship_get_title() validates
 * the statement.
 *
 * The 'title' of a relationship may be defined at create or edit time 
 * (in which case it is stored in the database as the title field and doesn't change much)
 * or,  if left blank,  it can be retrieved each time it's used
 * (which allows it to reflect changes in the title of the target,  but is
 * probably inefficient in most cases,  as it causes a node_load for all references).
 * Storing title in the triple table is an extension to the RDF model used internally,  
 * and is sort of experimental.
 *
 * @see relationship_save_statement() The serializer
 * @see relationship_set() Construct-and-save combined
 * @see relationship_array2statements() From turtle-like structure to triple structure
 *
 * @param $subject URI or node ID
 * @param $predicate shortname (or URI)
 * @param $object URI or Node ID
 * @param $title optional visible label for exposing this link
 * @param $why optional Something that explains why this statement was made, 
 *             eg an annotation by a user (uid), a deduced logic path (todo), or a system-defined fact (internal)
 * @return $statement assoc array structure.
 *
 */
function relationship_statement($subject,  $predicate,  $object,  $title=NULL,  $depth=NULL, $why=NULL) {
  $statement = array(
    'subject'     => (string)$subject,
    'predicate'   => (string)$predicate,
    'object'      => (string)$object,
    'title'       => (string)$title,
    'depth'       => (int)$depth,
    'why'         => ($why)?$why:'relationship_statement',
 );
 return $statement;
}


/** 
 * Return just the named column out of a collection of statements.
 *
 * A vertical slice from a horizontal, 2D array.
 * @todo this may be more elegant, but is it more efficient than 
 * using trivial named accessor functions like relationship_get_object() ?
 */
function relationship_get_column($statements,$column) {
  if(! $column){return $statements;}
  
  return array_map(
    create_function('$statement', 'return $statement["'.$column.'"];') ,
    $statements
  );
}
/**
 * Return a named array, pairing the contents of column a as the key for
 * the contents of column b.
 *
 * There may be a php function I'm missing that does this better.
 */
function relationship_get_assoc_column($statements, $columna, $columnb, $as_shortnames) {
  $a = array();
  foreach($statements as $s){
    $key = $as_shortnames ? relationship_get_shortname( $s[$columna] ) : $s[$columna] ;
    $a[$key] = $s[$columnb];
  }
  return $a;
}

/**
 * This actually converts statements made using fully qualified names
 * to our local shortname identifiers.
 *
 * This may be a little short-sighted, but it makes editing a lot
 * easier, and can be turned off here when we want to go to
 * full namespaces.
 * Transforms both predicate and object (but not subject) into canonical
 * shortnames from URLs (if the identifier can be found in the DB)
 *
 * @param a statement, passed and modified by ref
 * @return false if statement is invalid in some way
 * @see relationship_validate_statement()
 */
function relationship_validate_statement(&$s) {
#  $s['predicate'] = relationship_get_shortname($s['predicate']);
  if( $s['predicate'] != PREDICATE_NAME){   
    $s['object'] = relationship_get_shortname($s['object']);
  }
//  if( (! is_string($s['subject'])) || (! is_string($s['predicate'])) || (! is_string($s['object'])) ){
//    debug('Invalid statement, '._relationship_hash_key($s)." some value is not a string. (numbers must be cast to string before saving, objects and arrays mean somethings wrong)" ,0);
//    return FALSE;
//  }
  return $s;
}

/**
 * Save a given statement triple in the datastore.
 *
 * @param $statement 
 *        A statement is an assoc array defining 'subject',  'predicate',  'object',
 *        and optionally 'title' and 'uid' @see relationship_statement()
 * @see relationship_statement()
 * @see relationship_query()
 * @todo dupe checking
 * @todo datastore interface
 */
function relationship_save_statement($statement) {
  $ds = _relationship_datastore();
  $ds->add($statement);
}

/**
 * Basically a shorthand way to create and save a statement in one go.
 * @see relationship_statement()
 * @see relationship_save_statement()
 */
function relationship_set($subject,  $predicate,  $object,  $title=NULL,  $uid=NULL) {
  return relationship_save_statement(
    relationship_statement($subject,  $predicate,  $object,  $title,  $uid)
  );
}


/**
 * Attempt to make a SPARQL-type query on the datastore.
 *
 * relationship_query() does not automatically get titles like
 * relationship_get_statement() does. 
 * This is because it works on bigger sets and may not always need to display
 * the results.
 * If you wish to link to the results of a query, call relationship_get_titles()
 * on the result yourself before proceeding.
 *
 * @param $subject optional
 * @param $predicate optional
 * @param $object optional
 * @param $column optional
 * @returns a collection of statements
 */
function relationship_query($subject=NULL, $predicate=NULL, $object=NULL, $column=NULL) {

  // Allow the actual subject resource object to be passed in by reference
  // (is this used much?)
  if (is_object($subject) ) {
    if ($subject->uri) {
      $subject=$subject->uri;
    } else if ($subject->nid) {
      $subject=$subject->nid;
    }
  }

  debug("Querying "._relationship_hash_key(relationship_statement($subject,$predicate, $object)) , 2);


  // Informative paranoia is vital in an API
  // Do some assertions

  // Initialize, and check our datastore
  $ds = _relationship_datastore();

  if(! $subject_statements = $ds->query($subject,$predicate, $object)){
    $subject_statements = array();
  }
 
  // alias (sameAs) support?
 
 
 
  // If the datastore doesn't do it's own inverse support, I'll do it myself
  // note, I don't cache or anything
  if(! $ds->supported_characteristics(PREDICATE_INVERSE_OF) ){
    $query = relationship_statement($subject, $predicate, $object);
    $inverse_statements = relationship_query_inverseOf($query);
    $subject_statements = array_merge($inverse_statements,$subject_statements);
  }
 

  ////////////////////////
  // See if there is a custom hook for executing this query supplied by a
  // module or library.
  // If so, it can do all the work. 
  // Code will take precedence over any database lookups.
  // todo define this hook interface more
  // - should always return an array, even if empty
  ////////////////////////
  $custom_lookup = "relationship_lookup_".$predicate;
  if ( $predicate && function_exists($custom_lookup) ) {
    // A pluggable ontology has defined its own 
    // lookup routine,  let it take over from here.

    debug("Calling custom lookup function $custom_lookup($subject) ", 1);
    $custom_statements = $custom_lookup($subject,  $max_recursion_level,  $depth);
    debug("Custom lookup $custom_lookup() returned  ".print_r($subject_statements, 1), 2);
    $subject_statements = array_merge($custom_statements,$subject_statements);

    // That'll do for now. Cache results in the current context and return!
    
    // Not saving deduced values in the database permanently... yet
    // But if we did, it would be:
    // $ds->add_array($subject_statements, $custom_lookup);
  }

  // Filter to just the results
  if($column && $subject_statements){return relationship_get_column($subject_statements,$column);}

  return ($subject_statements) ? $subject_statements : array();
}


/**
 * Remove associations of a resource (node) to its terms.
 *
 * Used when updating node statements to start afresh.
 *
 * Optionally, also delete all references to it - this should only
 * be done when deleting an item and forgetting it ever existed.
 *
 * $param $subject nid, path or URI
 * $param $extreme_prejudice Wipe out anyone who ever came into contact with this ...
 */
function relationship_resource_delete_statements($subject, $extreme_prejudice=FALSE) {

  debug(t('Deleting links and statements about '.$subject),2);
  $ds = _relationship_datastore();
  $ds->delete($subject);  
  
  if($extreme_prejudice){
    if($rows_changed = $ds->delete(NULL,NULL,$subject) ){
      debug(t("$rows_changed statements REFERRING TO $subject also deleted."),2);  
    };
  }
}



/**
 * Return a handle on the datastore
 */
function _relationship_datastore() {
  static $ds;  
  if (!$ds) {$ds = new relationship_datastore();}
  return $ds;
}

/**
 * Lookup statements that imply the given query via the inverse rule.
 *
 * Predicate may be blank, which would mean anything that says anything about the subject.
 */
function relationship_query_inverseOf($query){
  #if(!$query){debug('no query? this is a waste of time');return;}
  
  $statements = array();
  $query = relationship_inverse_statement($query);
  if($query){
    $ds = _relationship_datastore();
    $inverse_statements = $ds->query($query);
    if($inverse_statements){
      debug("Found inverse statements: ".join(', ',array_keys($inverse_statements)),2);
      foreach($inverse_statements as $i => $statement){
        $s = relationship_inverse_statement($statement);
        $statements[_relationship_hash_key($s)] = $s;
      }
    }
  } else {
    debug("There is no inverse to this predicate. ".$query['predicate'] ,2);
  }
  
  return $statements;
}


/**
 * Reverse the operands and inverse the predicate for a statement.
 *
 * Eg turn a IsChildOf B into B HasChild A
 * If you already know the reciprocal,  pass it in,
 * otherwise it will be retrieved.
 *
 * Helper for apply_characteristic_inverseOf()
 *
 * Inverse of a statement with a blank predicate returns as an inverse
 * statement with a blank predicate. This is (almost) meaningless as a 
 * statement, but useful as a query.
 * The inverse of the question
 * {this, ?, ?}  - tell me everything about this
 * is
 * {?, ?, this}  - tell me everything that MENTIONS this
 *
 * @param $statement @see statements
 * @param $reciprocal predicate term definition @see predicates
 * @return modified statement,  also changes it by reference. 
 *         Returns null if no inverse predicate exists
 */
function relationship_inverse_statement(&$statement, $reciprocal='') {
  if (! $reciprocal && $statement['predicate']) {
    $term_def = relationship_get_term($statement['predicate']);
    if(!$term_def){ return; } // no such relationship
    $reciprocal = $term_def->inverseOf;
  }
    
  if ($reciprocal || (!$statement['predicate']) ) {
    $recip_statement = array(
      'subject'   => $statement['object'],
      'predicate' => $reciprocal,
      'object'    => $statement['subject'],  
      'inverseOf' => $statement['predicate'],  
      'title'     => '',
      'depth'     => $statement['depth'] ++,
      'why'       => 'inverse of '.$statement['predicate']
   );
  }
  return $statement = $recip_statement;
}





/**
 * Convert an assoc array to a collection of statements
 * about something.
 *
 * From a php properties array (very much like turtle syntax) :
 *
 *   $thisItem = array(
 *        'name' => 'joe',
 *        'age' => '55',
 *   );
 *
 * We learn that:
 *
 * $statements = array(
 *  [] = array(
 *      'subject' => 'thisItem',
 *      'predicate' => 'name',
 *      'object'    => 'joe',
 *    ),
 *  [] = array(
 *      'subject' => 'thisItem',
 *      'predicate' => 'age',
 *      'object'    => '55',
 *    ),
 *  );
 *
 * @see <http://www.dajobe.org/2004/01/turtle/>
 * @see relationship_statements2array()
 * @todo rename this turtle2triples() or something?
 * @param $properties assoc array
 * @param $subject the 'subject' identifier that these statements are being
 *                 made about.
 *                 If not defined, either the 'uri' or the 'name' provided 
 *                 in the data will be used instead. (in that order)
 * @param $import_template an assoc array matching keys used in the array to predicate
 *                 terms - if any translation is needed.
 * @param $strict  ONLY analyse statements listed in the import template.
 *                 This is handy when abstacting a limited amount of info out
 *                 of a complex object (like a node)
 *                 If off, all properties will be converted to triples
 * @return unindexed array of statements
 */
function relationship_array2statements($properties, $subject='anon' , $import_template = array() , $strict = FALSE ) {
  $statements = array();

  if(! $subject){
    $subject = $properties['uri'] ? $properties['uri'] : $properties['name'];
  }
  
  $key_mapping = array_flip($import_template);

  debug("Importing data array ".print_r($properties,1)." Using renaming map ".print_r($key_mapping,1) ,3);
  
  foreach($properties as $predicate => $object){
    if(isset($key_mapping[$predicate])){
      // rename certain array keys to our custom terms
      $predicate = $key_mapping[$predicate];
    } else if ($strict){
      continue;
    }
    // tidy the $object string (may have whitespace from the code layout)
    $object = preg_replace("/  +/"," ",$object);
    
    if($predicate){

      if($object && is_string($object)){
        $s = relationship_statement($subject,  $predicate,  $object, NULL, 0, 'internal');
        $statements[_relationship_hash_key($s)] = $s;
      } 
      else if(is_array($object)){
        // multiple values to one predicate are also legal shorthand
        // 'Robert' = array( 'sameAs' => array('Bob','Bobby') ) ;
        // gets broken into two statements.
        foreach($object as $ob){
          $s = relationship_statement($subject,  $predicate,  $ob, NULL, 0, 'internal');
          $statements[_relationship_hash_key($s)] = $s;
        }
      }

    }
  }
  return $statements;
}



/**
 * Given a load of statements, build a picture about the subject.
 *
 * Cast all the given relationships and data as array properties.
 * Collates a 'thing' from a loose collection of assertions.
 * Used for quick n dirty instantiation of a data cloud to something I can use.
 *
 * @param $statements a collection of statements, all about one subject.
 * @param $property_map an array indexing any renaming of predicates to properties that may be required.
 * @return an array of structured, indexed information
 * @see relationship_array2statements()
 * @ingroup relationship_api
 */
function relationship_statements2array($statements, $property_map=array(), $and_statements=FALSE) {
  if(! is_array($statements)){return array();}
  
  $thing = array();
  
  // debug("Translating from statements <pre>{".join('. ',array_keys($statements)) ."}</pre> to turtle array using name mapping rules <pre>".print_r($property_map,1)."</pre>",0);
  
  foreach ($statements as $i => $statement){
    $p = $statement['predicate'];
    if(isset($property_map[$p])){
      // change its name according to the rules.
      
      if(is_string($property_map[$p])){
        // set it as a string - but what if it's already set?
        
        $thing[$property_map[$p]] = $statement['object'];
      }
      else if(is_array($property_map[$p])){
        // it's a collection of similar statements
        $thing[$property_map[$p]][] = $statement['object'];
      }
      else {
        debug("Assigning a non-string property? Trying to set {$property_map[$p]} to {$statement['object']} could be trouble");
        $thing[$property_map[$p]] = $statement['object'];
      }

    } else {
      // this property is not in the naming map

      if(! $thing[$p] ){
        $thing[$p] = $statement['object'];
      } else {
        // auto-extension of properties into an array when a dupe is found.
        // This may casuse confusion down the track. Just check your strings
        // before working with them, they may be arrays!

        if (is_array($thing[$p])){
          $thing[$p][] = $statement['object'];
        } else {
          $old = $thing[$p];
          $thing[$p] = array($old,$statement['object']);
        }
        
      }
      if($and_statements){
        // retain the explicit statements in the object .. but only for
        // arbitrary facts, not built-ins.
        $thing['statements'][$i] = $statement;
      }
    } // not special
    
  } // each

  if($statement['subject']){
    $thing['uri'] = $statement['subject'];
  }
#debug_pre($thing);

  return $thing;
}

/**
 * @}
 */

?>
